# -*- coding: utf-8 -*-
# Copyright 2016-2023 tkwant authors.
#
# This file is part of tkwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# https://tkwant.kwant-project.org/doc/stable/pre/license.html.
# A list of tkwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# https://tkwant.kwant-project.org/doc/stable/pre/authors.html.

"""Test module for `tkwant.manybody`"""

import pytest
from pytest import raises
import cmath
import numpy as np
import kwant
from numpy.testing import (assert_array_almost_equal, assert_almost_equal)
from scipy.special import erf
import functools
import itertools
import copy
import tinyarray as ta
from math import cos, sin, exp


from .. import onebody, manybody, leads
# from tkwant.tests.common import (make_square, make_square_with_leads, make_simple_lead,
#                                  make_complex_lead, check_boundary_hamiltonian)

import kwantspectrum

import warnings
warnings.filterwarnings("ignore", category=DeprecationWarning)


def make_simple_lead(W=3):
    lat = kwant.lattice.square(1, norbs=1)
    sym = kwant.TranslationalSymmetry((-1, 0))
    H = kwant.Builder(sym)
    H[(lat(0, y) for y in range(W))] = 0
    H[lat.neighbors()] = -1
    return H


def onsite_potential(site, time, v):
    return 4 * (1 + v * cmath.exp(- 1j * time))


def gaussian(time, t0, A, sigma):
    return A * (1 + erf((time - t0) / sigma))


# time dependent coupling with gaussian pulse
def coupling_nn(site1, site2, time, t0, A, sigma):
    return - cmath.exp(- 1j * gaussian(time, t0, A, sigma))


def make_system(a=1, gamma=1.0, W=1, L=30):

    lat = kwant.lattice.square(a=a, norbs=1)
    syst = kwant.Builder()

    # Define the scattering region.
    syst[(lat(x, y) for x in range(L) for y in range(W))] = onsite_potential
    syst[lat.neighbors()] = -gamma
    # time dependent coupling between two sites 0 and 1
    for y in range(W):
        syst[lat(0, y), lat(1, y)] = coupling_nn

    # Define and attach the leads.
    # Construct the left lead.
    lead = kwant.Builder(kwant.TranslationalSymmetry((-a, 0)))
    lead[(lat(0, j) for j in range(W))] = 4 * gamma
    lead[lat.neighbors()] = -gamma

    # Attach the left lead and its reversed copy.
    syst.attach_lead(lead)
    syst.attach_lead(lead.reversed())

    return syst


def make_td_system(lat, N, td_onsite):
    I0 = ta.identity(2)
    syst = kwant.Builder()
    square = itertools.product(range(N), range(N))
    syst[(lat(i, j) for i, j in square)] = td_onsite
    syst[lat.neighbors()] = -1 * I0
    return syst


def make_td_system_with_leads(lat, N, lead_maker, td_onsite):
    syst = make_td_system(lat, N, td_onsite)
    syst.attach_lead(lead_maker(lat, N))
    syst.attach_lead(lead_maker(lat, N).reversed())
    return syst


def make_system_with_boundstate(V, L=20, a=1):  # boundstate present for V < 0

    lat = kwant.lattice.square(a=a, norbs=1)
    syst = kwant.Builder()

    # central system
    syst[(lat(i, 0) for i in range(-L, L))] = 0
    syst[lat(0, 0)] = V
    syst[lat.neighbors()] = -1

    # leads
    lead = kwant.Builder(kwant.TranslationalSymmetry((-a, 0)))
    lead[lat(0, 0)] = 0
    lead[lat.neighbors()] = -1
    syst.attach_lead(lead)
    syst.attach_lead(lead.reversed())

    return syst


def make_single_spin_rotating_basis(a, J, w, theta):

    # define pauli matrices
    sigma_0 = ta.array([[1, 0], [0, 1]])
    sigma_x = ta.array([[0, 1], [1 ,0]])
    sigma_z = ta.array([[1,0], [0, -1]])

    # central region
    lat = kwant.lattice.square(a=a, norbs=2)    # two orbitals for up and down spin
    syst = kwant.Builder()

    syst[lat(0, 0)]  = - (w / 2) * sigma_z - (J / 2) * (np.sin(theta) * sigma_x + np.cos(theta) * sigma_z)
    syst[lat(-1, 0)] = - (w / 2) * sigma_z  # add extra site to measure current
    
    syst[lat.neighbors()] = - sigma_0

    # leads
    lead=kwant.Builder(kwant.TranslationalSymmetry((-a, 0)))
    lead[lat(0, 0)] = - (w / 2) * sigma_z
    lead[lat.neighbors()] = - sigma_0
    syst.attach_lead(lead)
    syst.attach_lead(lead.reversed())
    
    return syst, lat



@pytest.fixture
def two_band_spectrum():
    lead = make_simple_lead(W=2).finalized()
    return kwantspectrum.spectrum(lead)


@pytest.fixture
def flat_band_spectrum():

    def make_lead_with_flat_band(ll=3):
        lat = kwant.lattice.square(norbs=1)
        lead = kwant.Builder(kwant.TranslationalSymmetry((-1, 1)))
        lead[[lat(0, j) for j in range(ll)]] = 0
        lead[lat.neighbors()] = -1
        return lead

    lead = make_lead_with_flat_band().finalized()
    return kwantspectrum.spectrum(lead)


@pytest.fixture
def uniform_occupation():
    def uniform_distribution(x):
        return np.ones(len(x))

    class UniformOccupation:
        def __init__(self):
            self.distribution = uniform_distribution
            self.bands = None
            self.energy_range = None
            self.multivariate = False
    return UniformOccupation()


@pytest.fixture
def SimpleInterval():

    class Interval:
        def __init__(self, kmin, kmax):
            self.kmin = kmin
            self.kmax = kmax
            self.checkattribute = 'check'  # dummy attribute to check if it is copied

        def __eq__(self, other):
            return (self.kmin == other.kmin and self.kmax == other.kmax and
                    self.checkattribute == other.checkattribute)

        def __ne__(self, other):
            return not self.__eq__(other)
    return Interval


@pytest.fixture
def quadrature_interval():
    return manybody.Interval(lead=0, band=0, kmin=0, kmax=np.pi / 2)


VALID_CHEMICAL_POTENTIALS = [-float("inf"), -1., -1, 0, 0., 1, 1., float("inf")]
VALID_TEMPERATURES = [0, 0., 1, 1., float("inf")]
VALID_ENERGY_CUTOFFS = [-float("inf"), -1., -1, 0, 0., 1, 1., float("inf")]

VALID_QUADRATURES = ["gausslegendre", "kronrod"]
VALID_INTEGRATION_VARIABLES = ["energy", "momentum"]


def unknown_distribution_function(chemical_potential, temperature, energy):
    """an arbitrary distribution function (here the fermi function)"""
    return 1 / (np.exp((energy - chemical_potential) / temperature) - 1)


def test_lead_occupation_default():
    occupation = manybody.lead_occupation()
    assert callable(occupation.distribution)
    assert occupation.distribution(-1) == 1
    assert occupation.distribution(1) == 0
    assert len(occupation.energy_range) == 1
    emin, emax = occupation.energy_range[0]
    assert emin is None
    assert emax == 0
    assert occupation.bands is None


def model_tasks():

    # we take a simple spectrum with only two bands.
    # it has the dispersion E_n(k), where n is the band index and k the momentum:
    # E_0(k) = - (2 cos(k) + 1)
    # E_1(k) = - (2 cos(k) - 1)

    def energy_func(momentum):
        return - (2 * np.cos(momentum) + 1)

    def velocity_func(momentum):
        return 2 * np.sin(momentum)

    def distribution_func(energy):
        return 1 / (np.exp((energy - 1) / 0.5) - 1)

    interval_1 = manybody.Interval(lead=0, band=0, kmin=0, kmax=np.pi / 2,
                                   integration_variable='momentum', order=4,
                                   quadrature='gausslegendre')

    # gauss legendre abscissas between kmin = 0 and kmax = pi/2
    momenta = [0.10906329, 0.51837768, 1.05241865, 1.46173304]
    weights_1 = [0.27320456, 0.51219361, 0.51219361, 0.27320456]

    interval_2 = manybody.Interval(lead=0, band=0, kmin=0, kmax=np.pi / 2,
                                   integration_variable='energy', order=4,
                                   quadrature='gausslegendre')

    # gauss legendre abscissas between E_0(kmin) = -3 and E_0(kmax) = -1
    energies = [-2.86113631, -2.33998104, -1.66001896, -1.13886369]
    weights_2 = [0.34785485, 0.65214515, 0.65214515, 0.34785485]

    variables = ('energy_func, velocity_func, distribution_func,'
                 'interval, weights, x, x_type')
    model = []
    model.append((energy_func, velocity_func, distribution_func,
                  interval_1, weights_1, momenta, 'momentum'))
    model.append((energy_func, velocity_func, distribution_func,
                  interval_2, weights_2, energies, 'energy'))

    return pytest.mark.parametrize(variables, model)


@pytest.mark.parametrize('chemical_potential', VALID_CHEMICAL_POTENTIALS)
def test_lead_occupation_vary_chemical_potential(chemical_potential):
    # zero temperature, fermi-dirac distribution
    occupation = manybody.lead_occupation(chemical_potential=chemical_potential)
    assert len(occupation.energy_range) == 1
    emin, emax = occupation.energy_range[0]
    assert emin is None
    assert emax is chemical_potential

    # finite temperature, fermi-dirac distribution
    occupation = manybody.lead_occupation(chemical_potential=chemical_potential, temperature=1)
    effective_emax = chemical_potential + manybody._fermi_function_below_epsilon(temperature=1)
    assert len(occupation.energy_range) == 1
    emin, emax = occupation.energy_range[0]
    assert emin is None
    assert emax == effective_emax


@pytest.mark.parametrize('temperature', VALID_TEMPERATURES)
def test_lead_occupation_vary_temperature(temperature):
    # TODO: no good test for prebind temperature value at the moment

    # fermi-dirac distribution
    occupation = manybody.lead_occupation(temperature=temperature)
    assert len(occupation.energy_range) == 1
    emin, emax = occupation.energy_range[0]
    if temperature > 0:
        # TODO: check
        assert emax == manybody._fermi_function_below_epsilon(temperature=temperature)
    else:
        assert emax == 0
    assert emin is None
    assert callable(occupation.distribution)


def test_lead_occupation_fermi_diract_zero_temperature():
    # check default values, finite temperature
    mu = 2
    epsilon = 1E-6  # some small number
    occupation = manybody.lead_occupation(chemical_potential=mu)
    _emin, _emax = occupation.energy_range[0]
    assert occupation.distribution(energy=mu - epsilon) == 1
    assert occupation.distribution(energy=mu + epsilon) == 0
    assert _emin is None
    assert _emax == mu
    assert occupation.bands is None


def test_lead_occupation_fermi_diract_finite_temperature():
    # check default values, finite temperature
    def fermi_function(chemical_potential, temperature, energy):
        return 1 / (np.exp((energy - chemical_potential) / temperature) + 1)

    mu = 2
    temp = 3
    occupation = manybody.lead_occupation(chemical_potential=mu, temperature=temp)
    _emin, _emax = occupation.energy_range[0]
    assert occupation.distribution(energy=1) == fermi_function(mu, temp, energy=1)
    assert occupation.distribution(energy=3) == fermi_function(mu, temp, energy=3)
    assert _emin is None
    assert _emax == mu + manybody._fermi_function_below_epsilon(temperature=temp)
    assert occupation.bands is None


def test_lead_occupation_band_selection():
    # check band selection
    occupation = manybody.lead_occupation(bands=3)
    assert occupation.bands == [3]
    occupation = manybody.lead_occupation(bands=[3])
    assert occupation.bands == [3]
    # bands always sorted and without dublicates
    occupation = manybody.lead_occupation(bands=[3, 2, 5, 5, -1])
    assert occupation.bands == [-1, 2, 3, 5]


def test_lead_occupation_own_datatype():
    class MyOccupation:
        def __init__(self, distribution, energy_range, bands, multivariate):
            self.distribution = distribution
            self.bands = bands
            self.energy_range = energy_range
            self.multivariate = multivariate
            self.somefield = 42
    occupation = manybody.lead_occupation(occupation_type=MyOccupation)
    assert callable(occupation.distribution)
    assert occupation.bands is None
    assert occupation.energy_range == [(None, 0)]
    assert occupation.multivariate is False
    assert occupation.somefield == 42


def test_lead_occupation_raises():
    # negative temperature
    raises(ValueError, manybody.lead_occupation, temperature=-1)
    # not real numbers
    raises(TypeError, manybody.lead_occupation, chemical_potential=1 + 1j)
    raises(TypeError, manybody.lead_occupation, temperature=1 + 1j)


def test_make_intervals_default(two_band_spectrum, uniform_occupation):
    intervals = manybody.calc_intervals(two_band_spectrum, uniform_occupation)

    # check all types of the returned elements
    assert isinstance(intervals, list)
    assert len(intervals) == two_band_spectrum.nbands
    for interval in intervals:
        assert isinstance(interval, manybody.Interval)
        assert isinstance(interval.lead, int)
        assert isinstance(interval.band, int)
        assert isinstance(interval.kmin, float)
        assert isinstance(interval.kmax, float)
        assert isinstance(interval.order, int)
        assert isinstance(interval.integration_variable, str)
        assert isinstance(interval.quadrature, str)

    # check explicit returned values
    # intervals for all bands of the spectrum
    # intervals contains the momentum (k) intervals where d E_n / dk > 0
    # from above dispersion follows that this is the case for k in [0, pi]
    for i, interval in enumerate(intervals):
        assert interval.lead == 0  # only one lead
        assert interval.band == i
        assert_almost_equal(interval.kmin, 0)
        assert_almost_equal(interval.kmax, np.pi)
        assert interval.order == 10
        assert interval.integration_variable == 'momentum'
        assert interval.quadrature == 'kronrod'


def test_make_intervals_with_unoccupied_leads(two_band_spectrum, uniform_occupation):
    # check that intervals with additional unoccupied leads is similar
    intervals = manybody.calc_intervals(two_band_spectrum, uniform_occupation)

    for nb_leads in range(1, 5):
        spectra = [two_band_spectrum] * nb_leads
        occupations = [uniform_occupation] + [None] * (nb_leads - 1)
        intervals_unocc = manybody.calc_intervals(spectra, occupations)
        assert intervals_unocc == intervals


def test_make_intervals_lead_index(two_band_spectrum, uniform_occupation):
    # check that the position of the occupation determines the lead index
    total_nb_leads = 5
    for i in range(0, total_nb_leads):
        spectra = [two_band_spectrum] * total_nb_leads
        occupations = [None] * i + [uniform_occupation] + [None] * (total_nb_leads - i - 1)
        intervals = manybody.calc_intervals(spectra, occupations)
        for interval in intervals:
            assert interval.lead == i


def test_make_intervals_with_energy_constraint(two_band_spectrum, uniform_occupation):

    # select only bands in energy range emin < E_n(k) < emax
    # intervals contains the momentum (k) intervals where d E_n / dk > 0
    # with the additional constraint emin < E_n(k) < emax
    uniform_occupation.energy_range = [(-1, 1)]
    intervals = manybody.calc_intervals(two_band_spectrum, uniform_occupation)

    assert len(intervals) == two_band_spectrum.nbands
    for i, interval in enumerate(intervals):
        assert isinstance(interval, manybody.Interval)
        assert interval.lead == 0  # only one lead
        assert interval.band == i
        if interval.band == 0:
            assert_almost_equal(interval.kmin, 0.5 * np.pi)
            assert_almost_equal(interval.kmax, np.pi)
        else:
            assert_almost_equal(interval.kmin, 0)
            assert_almost_equal(interval.kmax, 0.5 * np.pi)


def test_make_intervals_with_energy_and_band_constraint(two_band_spectrum, uniform_occupation):

    # select only bands in energy range and take only lowest band with index 0
    # intervals contains the momentum (k) intervals where d E_0 / dk > 0
    # with the additional constraint emin < E_0(k) < emax
    # (note that we have set n to 0)
    uniform_occupation.energy_range = [(-1, 1)]
    uniform_occupation.bands = [0]
    intervals = manybody.calc_intervals(two_band_spectrum, uniform_occupation)

    assert len(intervals) == 1
    for i, interval in enumerate(intervals):
        assert isinstance(interval, manybody.Interval)
        assert interval.lead == 0  # only one lead
        assert interval.band == i
        assert_almost_equal(interval.kmin, 0.5 * np.pi)
        assert_almost_equal(interval.kmax, np.pi)


def test_make_intervals_which_are_empty(two_band_spectrum, uniform_occupation):
    # case where no intervals are found
    uniform_occupation.energy_range = [(5, 6)]
    intervals = manybody.calc_intervals(two_band_spectrum, uniform_occupation)
    assert isinstance(intervals, list)
    assert len(intervals) == 0


def test_make_intervals_with_two_similar_occupied_leads(two_band_spectrum, uniform_occupation):
    # two leads with equal occupation
    num_leads = 2
    intervals = manybody.calc_intervals([two_band_spectrum] * num_leads, uniform_occupation)
    assert len(intervals) == two_band_spectrum.nbands * num_leads
    for i, interval in enumerate(intervals):
        assert isinstance(interval, manybody.Interval)
        assert interval.lead in range(num_leads)
        assert_almost_equal(interval.kmin, 0)
        assert_almost_equal(interval.kmax, np.pi)


def test_make_intervals_with_two_differently_occupied_leads(two_band_spectrum, uniform_occupation):
    # two leads with different occupations
    num_leads = 2
    occupation_1 = copy.deepcopy(uniform_occupation)
    occupation_1.bands = [0]
    occupation_2 = copy.deepcopy(uniform_occupation)
    occupation_2.energy_range = [(-1, 1)]
    occupation_2.bands = [0]
    intervals = manybody.calc_intervals([two_band_spectrum] * num_leads, [occupation_1, occupation_2])
    assert len(intervals) == num_leads  # only one band per lead
    for i, interval in enumerate(intervals):
        assert isinstance(interval, manybody.Interval)
        if interval.lead == 0:
            assert_almost_equal(interval.kmin, 0)
            assert_almost_equal(interval.kmax, np.pi)
        elif interval.lead == 1:
            assert_almost_equal(interval.kmin, 0.5 * np.pi)
            assert_almost_equal(interval.kmax, np.pi)


def test_make_intervals_interface(two_band_spectrum, uniform_occupation):
    # check for robustness of interfact for sequences input
    intervals_1 = manybody.calc_intervals(two_band_spectrum, uniform_occupation)
    intervals_2 = manybody.calc_intervals([two_band_spectrum], uniform_occupation)
    intervals_3 = manybody.calc_intervals(two_band_spectrum, [uniform_occupation])
    intervals_4 = manybody.calc_intervals([two_band_spectrum], [uniform_occupation])
    assert intervals_1 == intervals_2 == intervals_3 == intervals_4
    intervals_1 = manybody.calc_intervals([two_band_spectrum] * 2, [uniform_occupation] * 2)
    intervals_2 = manybody.calc_intervals((two_band_spectrum, two_band_spectrum), uniform_occupation)
    assert intervals_1 == intervals_2  # only true in equal occupation


def test_make_intervals_band_index_out_of_range(two_band_spectrum, uniform_occupation):
    uniform_occupation.bands = [two_band_spectrum.nbands]
    raises(AssertionError, manybody.calc_intervals, two_band_spectrum, uniform_occupation)


def test_make_intervals_band_index_not_integer(two_band_spectrum, uniform_occupation):
    uniform_occupation.bands = [0.]
    raises(AssertionError, manybody.calc_intervals, two_band_spectrum, uniform_occupation)


def test_make_intervals_energies_interchanged(two_band_spectrum, uniform_occupation):
    uniform_occupation.energy_range = [(1, 0)]
    raises(AssertionError, manybody.calc_intervals, two_band_spectrum, uniform_occupation)


def test_make_intervals_emin_not_real(two_band_spectrum, uniform_occupation):
    uniform_occupation.energy_range = [(1 + 1j, None)]
    # todo
    # raises(TypeError, manybody.calc_intervals, two_band_spectrum, uniform_occupation)


def test_make_intervals_emax_not_real(two_band_spectrum, uniform_occupation):
    uniform_occupation.energy_range = [(None, 1 + 1j)]
    # todo
    # raises(TypeError, manybody.calc_intervals, two_band_spectrum, uniform_occupation)


def test_make_intervals_inconsitent_input(two_band_spectrum, uniform_occupation):
    # inhomogeneous occupation, that does not fit the number of leads
    raises(ValueError, manybody.calc_intervals, two_band_spectrum, [uniform_occupation] * 2)
    raises(ValueError, manybody.calc_intervals, [two_band_spectrum], [uniform_occupation] * 2)
    raises(ValueError, manybody.calc_intervals, [two_band_spectrum] * 3, [uniform_occupation] * 2)


def test_make_intervals_missing_energy_range_attribute(two_band_spectrum, uniform_occupation):
    del uniform_occupation.energy_range
    raises(AttributeError, manybody.calc_intervals, two_band_spectrum, uniform_occupation)


def test_make_intervals_missing_bands_attribute(two_band_spectrum, uniform_occupation):
    del uniform_occupation.bands
    raises(AttributeError, manybody.calc_intervals, two_band_spectrum, uniform_occupation)


def test_make_intervals_missing_distribution_attribute(two_band_spectrum, uniform_occupation):
    del uniform_occupation.distribution
    # todo
#    raises(AttributeError, manybody.calc_intervals, two_band_spectrum, uniform_occupation)


def test_split_interval_into_two(SimpleInterval):

    nint = 1  # number of returned intervals
    intervals = manybody._split_interval(interval=SimpleInterval(kmin=0, kmax=1),
                                         number_subintervals=nint)
    assert isinstance(intervals, list)
    assert len(intervals) == nint
    assert intervals == [SimpleInterval(kmin=0, kmax=1)]


def test_split_interval_into_three(SimpleInterval):

    nint = 2  # number of returned intervals
    intervals = manybody._split_interval(interval=SimpleInterval(kmin=0, kmax=1),
                                         number_subintervals=nint)
    assert isinstance(intervals, list)
    assert len(intervals) == nint
    assert intervals == [SimpleInterval(kmin=0, kmax=0.5),
                         SimpleInterval(kmin=0.5, kmax=1)]


def test_split_interval_wrong_input(SimpleInterval):
    raises(ValueError, manybody._split_interval, SimpleInterval(kmin=0, kmax=1), number_subintervals=0)
    raises(AssertionError, manybody._split_interval, SimpleInterval(kmin=0, kmax=1), number_subintervals=1.)


def test_split_interval_kmin_attribute_missing(SimpleInterval):
    interval = SimpleInterval(0, 1)
    del interval.kmin
    raises(AttributeError, manybody._split_interval, interval, 2)


def test_split_interval_kmax_attribute_missing(SimpleInterval):
    interval = SimpleInterval(0, 1)
    del interval.kmax
    raises(AttributeError, manybody._split_interval, interval, 2)


def test_split_intervals_no_split(SimpleInterval):
    nint = 1  # number of returned intervals per input interval
    input_intervals = [SimpleInterval(kmin=0, kmax=1), SimpleInterval(kmin=-4.0, kmax=2)]
    intervals = manybody.split_intervals(input_intervals, number_subintervals=nint)
    assert isinstance(intervals, list)
    assert len(intervals) == nint * len(input_intervals)
    assert intervals == [SimpleInterval(kmin=0, kmax=1), SimpleInterval(kmin=-4.0, kmax=2)]


def test_split_intervals_into_two(SimpleInterval):
    nint = 2  # number of returned intervals per input interval
    input_intervals = [SimpleInterval(kmin=0, kmax=1), SimpleInterval(kmin=-4.0, kmax=2)]
    intervals = manybody.split_intervals(input_intervals, number_subintervals=nint)
    assert isinstance(intervals, list)
    assert len(intervals) == nint * len(input_intervals)
    assert intervals == [SimpleInterval(kmin=0, kmax=0.5), SimpleInterval(kmin=0.5, kmax=1),
                         SimpleInterval(kmin=-4.0, kmax=-1.0), SimpleInterval(kmin=-1.0, kmax=2)]


def test_split_intervals_wrong_input(SimpleInterval):
    input_intervals = [SimpleInterval(kmin=0, kmax=1)]
    raises(ValueError, manybody.split_intervals, input_intervals, number_subintervals=0)
    raises(AssertionError, manybody.split_intervals, input_intervals, number_subintervals=1.)
    # interval not a list
    input_intervals = SimpleInterval(kmin=0, kmax=1)
    raises(TypeError, manybody.split_intervals, input_intervals, number_subintervals=2)


def test_attributes_similar():

    # use the manybody interval class to test for similarity
    int0 = manybody.Interval(lead=0, band=1, kmin=0, kmax=1.)
    int1 = manybody.Interval(lead=1, band=1, kmin=1E-12, kmax=1.)

    # check type
    res = manybody._attributes_similar(int0, int1, ['band', 'kmin', 'kmax'])
    assert isinstance(res, bool)

    # compare float attributes
    assert manybody._attributes_similar(int0, int1, ['band', 'kmin', 'kmax'])
    assert not manybody._attributes_similar(int0, int1, ['band', 'kmin', 'kmax'],
                                            rtol=0, atol=1E-15)

    # compare integer attributes
    assert not manybody._attributes_similar(int0, int1, ['lead', 'band', 'kmin', 'kmax'])

    # compare nonexisting attributes
    assert manybody._attributes_similar(int0, int1, ['bla'])

    # compare existing and nonexisting attributes
    int2 = manybody.Interval(lead=1, band=1, kmin=1E-12, kmax=1.)
    int2.bla = 42
    assert not manybody._attributes_similar(int0, int2, ['bla'])

    # compare float and numpy attributes
    int3 = manybody.Interval(lead=0, band=1, kmin=np.array([0.]), kmax=1.)
    assert manybody._attributes_similar(int1, int3, ['band', 'kmin', 'kmax'])
    assert not manybody._attributes_similar(int1, int3, ['band', 'kmin', 'kmax'],
                                            rtol=0, atol=1E-15)

    # compare numpy attributes
    int4 = manybody.Interval(lead=0, band=1, kmin=np.array([1E-12]), kmax=1.)
    assert manybody._attributes_similar(int3, int4, ['band', 'kmin', 'kmax'])
    assert not manybody._attributes_similar(int3, int4, ['band', 'kmin', 'kmax'],
                                            rtol=0, atol=1E-15)

    # compare numpy attributes with different sizes
    int5 = manybody.Interval(lead=0, band=1, kmin=np.array([0, 1E-12]), kmax=1.)
    assert manybody._attributes_similar(int4, int5, ['band', 'kmin', 'kmax'])


def test_combine_intervals():

    int0 = manybody.Interval(lead=3, band=1, kmin=0, kmax=1.)
    int1 = manybody.Interval(lead=5, band=1, kmin=1E-12, kmax=1.)
    int2 = manybody.Interval(lead=6, band=1, kmin=1E-12, kmax=1.)
    int3 = manybody.Interval(lead=7, band=1, kmin=0.1, kmax=1.)
    int4 = manybody.Interval(lead=(1, 2), band=1, kmin=0, kmax=1.)

    # combine two intervals
    combined = manybody.combine_intervals([int0, int1])
    assert isinstance(combined, list)
    assert len(combined) == 1
    assert combined[0].lead == (3, 5)
    assert combined[0].band == int0.band
    assert combined[0].kmin == int0.kmin
    assert combined[0].kmax == int0.kmax
    assert int0.lead == 3
    assert int1.lead == 5

    # do not combine intervals
    combined = manybody.combine_intervals([int0, int3])
    assert isinstance(combined, list)
    assert len(combined) == 2
    assert combined[0].lead == 3
    assert combined[1].lead == 7

    # combine an interval with lead integer and lead tuple
    combined = manybody.combine_intervals([int0, int4])
    assert len(combined) == 1
    assert combined[0].lead == (1, 2, 3)

    # combine on an empty list
    combined = manybody.combine_intervals([])
    assert isinstance(combined, list)
    assert len(combined) == 0

    # combine on a list with only one interval
    combined = manybody.combine_intervals([int0])
    assert isinstance(combined, list)
    assert len(combined) == 1
    assert combined[0].lead == 3

    # combine three intervals
    combined = manybody.combine_intervals([int0, int1, int2])
    assert isinstance(combined, list)
    assert len(combined) == 1
    assert combined[0].lead == (3, 5, 6)

    # combine two out of three intervals
    combined = manybody.combine_intervals([int0, int1, int3])
    assert len(combined) == 2
    assert combined[0].lead == (3, 5)
    assert combined[1].lead == 7

    combined = manybody.combine_intervals([int3, int1, int0])
    assert len(combined) == 2
    assert combined[0].lead == 7
    assert combined[1].lead == (3, 5)

    combined = manybody.combine_intervals([int1, int3, int0])
    assert len(combined) == 2
    assert combined[0].lead == (3, 5)
    assert combined[1].lead == 7

    # input list has some elements without lead attribute
    # if bla is the last arguement, nothing will be combined
    with pytest.raises(AttributeError) as exc:
        manybody.combine_intervals(['bla', int0])
    assert 'attribute "lead" missing in interval' in str(exc.value)

    with pytest.raises(AttributeError) as exc:
        manybody.combine_intervals([int0, 'bla', int1])
    assert 'attribute "lead" missing in interval' in str(exc.value)


def test_calc_modes_and_weights(two_band_spectrum, quadrature_interval):

    # check all types of the returned elements
    def distribution(energy):
        return np.ones(len(energy))

    quadrature_interval.quadrature = 'gausslegendre'

    tmp = manybody._calc_modes_and_weights(quadrature_interval, distribution, two_band_spectrum)
    modes, energies, momenta, weights, math_weights, phys_weights = tmp

    order = quadrature_interval.order

    # test shapes of the return elements

    assert isinstance(modes, np.ndarray)
    assert isinstance(energies, np.ndarray)
    assert isinstance(momenta, np.ndarray)
    assert isinstance(weights, np.ndarray)
    assert isinstance(math_weights, np.ndarray)
    assert isinstance(phys_weights, np.ndarray)

    assert modes.shape == (order,)
    assert energies.shape == (order,)
    assert momenta.shape == (order,)
    assert phys_weights.shape == (order,)

    assert weights.shape == (order,)
    assert math_weights.shape == (order,)

    quadrature_interval.quadrature = 'kronrod'

    tmp = manybody._calc_modes_and_weights(quadrature_interval, distribution, two_band_spectrum)
    modes, energies, momenta, weights, math_weights, phys_weights = tmp

    # test shapes of the return elements

    assert isinstance(modes, np.ndarray)
    assert isinstance(energies, np.ndarray)
    assert isinstance(momenta, np.ndarray)
    assert isinstance(weights, np.ndarray)
    assert isinstance(math_weights, np.ndarray)
    assert isinstance(phys_weights, np.ndarray)

    assert modes.shape == (2 * order + 1,)
    assert energies.shape == (2 * order + 1,)
    assert momenta.shape == (2 * order + 1,)
    assert phys_weights.shape == (2 * order + 1,)

    assert weights.shape == (2 * order + 1, 2)
    assert math_weights.shape == (2 * order + 1, 2)

    # no vector function distribution

    def fermi_dirac_distribution(energy):
        """Fermi diract distribution function"""
        return 1. / (1 + exp((energy - 0.5) / 0.1))
    distribution = fermi_dirac_distribution
    tmp = manybody._calc_modes_and_weights(quadrature_interval, distribution, two_band_spectrum)


@model_tasks()
def test_calc_tasks_values(two_band_spectrum, uniform_occupation, energy_func,
                           velocity_func, distribution_func,
                           interval, weights, x, x_type):

    uniform_occupation.distribution = distribution_func

    tasks = manybody.calc_tasks(interval, two_band_spectrum, uniform_occupation)

    if x_type == 'momentum':
        for i, k in enumerate(x):
            energy = energy_func(k)
            weight = distribution_func(energy) * velocity_func(k) * weights[i] / (2 * np.pi)
            assert_almost_equal(energy, tasks[i].energy)
            assert_almost_equal(weight, tasks[i].weight)
    elif x_type == 'energy':
        for i, energy in enumerate(x):
            weight = distribution_func(energy) * weights[i] / (2 * np.pi)
            assert_almost_equal(energy, tasks[i].energy)
            assert_almost_equal(weight, tasks[i].weight)
    else:
        raise NotImplementedError('type={} unknown'.format(x_type))


def test_calc_tasks_default(two_band_spectrum, uniform_occupation, quadrature_interval):

    # check all types of the returned elements
    tasks = manybody.calc_tasks(quadrature_interval, two_band_spectrum, uniform_occupation)

    assert isinstance(tasks, dict)
    assert len(tasks) == quadrature_interval.order * 2 + 1
    for i, (key, value) in enumerate(tasks.items()):

        assert key == i
        assert isinstance(value.lead, int)
        # assert isinstance(value.mode, int) # some numpy number, TODO
        assert isinstance(value.energy, float)
        # assert isinstance(value.weight, float) # for gausslegendre
        # assert isinstance(value.weight, np.ndarray) # for kronrod
        # TODO, make weight type similar


def test_calc_tasks_with_flat_band(flat_band_spectrum, uniform_occupation):

    # this is the interval with the flat band (velocity zero)
    interval = manybody.Interval(lead=0, band=1, kmin=-np.pi, kmax=np.pi)

    # check that no tasks are calculated as weights are zero
    tasks = manybody.calc_tasks(interval, flat_band_spectrum, uniform_occupation)

    assert len(tasks) == 0

    # check that we can tune the tolerance to get all tasks again
    tasks = manybody.calc_tasks(interval, flat_band_spectrum, uniform_occupation, tol=1E-30)

    assert len(tasks) == 2 * interval.order + 1


def test_calc_tasks_keys(two_band_spectrum, uniform_occupation, quadrature_interval):

    tasks = manybody.calc_tasks(quadrature_interval, two_band_spectrum, uniform_occupation, keys=itertools.count(5))

    for i, (key, task) in enumerate(tasks.items()):
        assert key == i + 5  # all keys shifted by offset


def test_calc_tasks_lead_index(two_band_spectrum, uniform_occupation, quadrature_interval):

    quadrature_interval.lead = 0
    tasks = manybody.calc_tasks(quadrature_interval, two_band_spectrum, uniform_occupation)

    for i, (key, task) in enumerate(tasks.items()):
        assert isinstance(task.lead, int)
        assert task.lead == 0


def test_calc_tasks_lead_index_tuple(two_band_spectrum, uniform_occupation):

    # calculate tasks with a sequence of intervals
    inter0 = manybody.Interval(lead=0, band=0, kmin=0, kmax=np.pi / 2)
    inter1 = manybody.Interval(lead=1, band=0, kmin=0, kmax=np.pi / 2)
    intervals_sep = [inter0, inter1]
    spectra = [two_band_spectrum, two_band_spectrum]

    tasks_sep = manybody.calc_tasks(intervals_sep, spectra, uniform_occupation)

    # calculate tasks with only one interval including both leads
    intervals_grp = manybody.Interval(lead=(0, 1), band=0, kmin=0, kmax=np.pi / 2)
    tasks_grp = manybody.calc_tasks(intervals_grp, spectra, uniform_occupation)

    assert len(tasks_sep) == len(tasks_grp)
    for key, value in tasks_sep.items():
        assert tasks_grp[key].lead == value.lead
        assert tasks_grp[key].mode == value.mode
        assert_almost_equal(tasks_grp[key].energy, value.energy)
        assert_almost_equal(tasks_grp[key].momentum, value.momentum)
        assert_array_almost_equal(tasks_grp[key].weight, value.weight)
        assert_array_almost_equal(tasks_grp[key].math_weight, value.math_weight)
        assert_array_almost_equal(tasks_grp[key].phys_weight, value.phys_weight)


def test_calc_tasks_band_index(two_band_spectrum, uniform_occupation, quadrature_interval):

    for band in range(two_band_spectrum.nbands):
        quadrature_interval.band = band
        tasks = manybody.calc_tasks(quadrature_interval, two_band_spectrum, uniform_occupation)
        assert isinstance(tasks, dict)
    # no specific test so far, as only numerical values change


@pytest.mark.parametrize('quadrature', VALID_QUADRATURES)
def test_calc_tasks_quadrature(two_band_spectrum, uniform_occupation, quadrature_interval, quadrature):

    quadrature_interval.quadrature = quadrature
    tasks = manybody.calc_tasks(quadrature_interval, two_band_spectrum, uniform_occupation)

    # todo: make weight type similar (numpy array)
    if quadrature == 'gausslegendre':
        assert len(tasks) == quadrature_interval.order
        for i, (key, task) in enumerate(tasks.items()):
            assert isinstance(task.lead, int)
            assert isinstance(task.weight, float)
    elif quadrature == 'kronrod':
        assert len(tasks) == 2 * quadrature_interval.order + 1
        for i, (key, task) in enumerate(tasks.items()):
            assert isinstance(task.lead, int)
            assert isinstance(task.weight, np.ndarray)
    else:
        raise NotImplementedError('quadrature={} unknown'.format(quadrature))


def test_calc_tasks_attribute_copy(two_band_spectrum, uniform_occupation, quadrature_interval):

    class MyTask:
        def __init__(self, weight, phys_weight, math_weight, energy, momentum, mode, lead):
            self.weight = weight
            self.phys_weight = phys_weight
            self.math_weight = math_weight
            self.energy = energy
            self.momentum = momentum
            self.mode = mode
            self.lead = lead
            self.checkattribute = 'check_task'  # dummy attribute to check if it is copied

    tasks = manybody.calc_tasks(quadrature_interval, two_band_spectrum, uniform_occupation, task_type=MyTask)

    for i, (key, task) in enumerate(tasks.items()):
        assert task.checkattribute == 'check_task'


@pytest.mark.parametrize('integration_variable', VALID_INTEGRATION_VARIABLES)
def test_calc_tasks_integration_variable(two_band_spectrum, uniform_occupation, quadrature_interval, integration_variable):

    quadrature_interval.integration_variable = integration_variable
    tasks = manybody.calc_tasks(quadrature_interval, two_band_spectrum, uniform_occupation)
    assert isinstance(tasks, dict)
    # no specific test so far, as only numerical values change


def test_calc_tasks_lead_index_out_of_range(quadrature_interval, two_band_spectrum,
                                            uniform_occupation):
    quadrature_interval.lead = 1
    raises(IndexError, manybody.calc_tasks, quadrature_interval, two_band_spectrum, uniform_occupation)


def test_calc_tasks_band_index_out_of_range(quadrature_interval, two_band_spectrum,
                                            uniform_occupation):
    quadrature_interval.band = 2
    raises(AssertionError, manybody.calc_tasks, quadrature_interval, two_band_spectrum, uniform_occupation)


def test_calc_tasks_kmin_kmax_interchanged(quadrature_interval, two_band_spectrum,
                                           uniform_occupation):
    quadrature_interval.kmin = 1
    quadrature_interval.kmax = 0
    raises(ValueError, manybody.calc_tasks, quadrature_interval, two_band_spectrum, uniform_occupation)


def test_calc_tasks_order_out_of_range(quadrature_interval, two_band_spectrum,
                                       uniform_occupation):
    quadrature_interval.order = 0
    raises(ValueError, manybody.calc_tasks, quadrature_interval, two_band_spectrum, uniform_occupation)


def test_calc_tasks_nonexisting_quadrature(quadrature_interval, two_band_spectrum,
                                           uniform_occupation):
    quadrature_interval.quadrature = 'bla'
    raises(NotImplementedError, manybody.calc_tasks, quadrature_interval, two_band_spectrum, uniform_occupation)


def test_calc_tasks_nonexisting_integration_variable(quadrature_interval, two_band_spectrum,
                                                     uniform_occupation):
    quadrature_interval.integration_variable = 'bla'
    raises(NotImplementedError, manybody.calc_tasks, quadrature_interval, two_band_spectrum, uniform_occupation)


def test_calc_tasks_distribution_not_callable(quadrature_interval, two_band_spectrum,
                                              uniform_occupation):
    uniform_occupation.distribution = 1
    raises(TypeError, manybody.calc_tasks, quadrature_interval, two_band_spectrum, uniform_occupation)


@pytest.mark.integtest  # marker for integration tests
def test_calc_initial_state():

    # ---- check initial state
    chemical_potential = 3
    tmax = 500
    params = {'v': 0.1, 't0': 0, 'A': 0.00157, 'sigma': 24}
    syst = make_system().finalized()
    occupation = manybody.lead_occupation(chemical_potential)
    spectrum = kwantspectrum.spectra(syst.leads)
    intervals = manybody.calc_intervals(spectrum, occupation)
    intervals[0].order = 10
    intervals[1].order = 8
    for interval in intervals:  # use simple quadrature
        interval.quadrature = 'gausslegendre'
    tasks = manybody.calc_tasks(intervals, spectrum, occupation)
    boundaries = leads.automatic_boundary(spectrum, tmax)
    psi_init = manybody.calc_initial_state(syst, tasks, boundaries,
                                           params=params)

    # check all types of the returned elements
    assert isinstance(psi_init, dict)
    assert len(psi_init) == intervals[0].order + intervals[1].order
    for i, (key, value) in enumerate(psi_init.items()):
        assert i == key
        assert isinstance(value, onebody.WaveFunction)


@pytest.mark.integtest
def test_calc_initial_state_with_kwant_failing():

    syst = make_system().finalized()
    occupation = manybody.lead_occupation(chemical_potential=3)
    spectrum = kwantspectrum.spectra(syst.leads)
    intervals = manybody.calc_intervals(spectrum, occupation)
    tasks = manybody.calc_tasks(intervals, spectrum, occupation)
    boundaries = [leads.SimpleBoundary(num_buffer_cells=2)] * len(syst.leads)
    params = {'v': 0.1, 't0': 0, 'A': 0.00157, 'sigma': 24}

    class ScatteringStatesRaise:
        """raise with a runtime error"""
        def __init__(self, *args, wavefunction_type=onebody.WaveFunction.from_kwant, **kwargs):
            raise RuntimeError()

    # we just mock that state calculation fails
    with pytest.raises(RuntimeError) as exc:
        manybody.calc_initial_state(syst, tasks, boundaries, params=params,
                                    scattering_state_type=ScatteringStatesRaise)
    # we do not check for the exact energy value as this can change
    assert 'scattering state calculation failed for energy=' in str(exc.value)


@pytest.mark.integtest  # marker for integration tests
def test_manybody_wavefunction():
    syst = make_system().finalized()

    density_operator = kwant.operator.Density(syst)  # array
    sum_density_operator = kwant.operator.Density(syst, sum=True)  # only one element in array

    chemical_potential = 3
    tmax = 500
    time = 5
    params = {'v': 0.1, 't0': 0, 'A': 0.00157, 'sigma': 24}

    occupation = manybody.lead_occupation(chemical_potential)
    spectrum = kwantspectrum.spectra(syst.leads)
    intervals = manybody.calc_intervals(spectrum, occupation)
    for interval in intervals:  # use simple quadrature
        interval.quadrature = 'gausslegendre'
    tasks = manybody.calc_tasks(intervals, spectrum, occupation)
    boundaries = leads.automatic_boundary(spectrum, tmax)
    psi_init = manybody.calc_initial_state(syst, tasks, boundaries, params=params)

    state = manybody.WaveFunction(psi_init, tasks)
    state.evolve(time)
    assert state.time == time
    for _, psi in state.psi._data.items():
        assert psi.time == time
        assert psi.params == {'v': 0.1, 't0': 0, 'A': 0.00157, 'sigma': 24}

    # check return type, we expect an array running over all sites or just a single number
    density = state.evaluate(density_operator)
    sum_density = state.evaluate(sum_density_operator)
    assert isinstance(density, np.ndarray)
    assert density.shape == (len(syst.sites), )
    assert isinstance(sum_density, np.ndarray)
    assert sum_density.shape == ()

    # check that expectation value for onebody/manybody wavefunction are similar
    onebody_density = state.get_onebody_state(key=0).evaluate(density_operator)
    onebody_sum_density = state.get_onebody_state(key=0).evaluate(sum_density_operator)

    assert density.shape == onebody_density.shape
    assert sum_density.shape == onebody_sum_density.shape

    # TODO: check quadrature also for kronrod like rules

    # get all keys
    keys = state.get_keys()
    assert isinstance(keys, list)
    assert len(keys) == len(tasks)
    assert np.array([isinstance(x, int) for x in keys]).all()

    # get a new free key
    key = state.get_free_key()
    assert isinstance(key, int)
    assert key not in state.get_keys()

    # add a new onebody state
    class Task:
        weight = 0.5
    task = Task()
    psi = onebody.ScatteringStates(syst, lead=0, energy=2.4, tmax=tmax, params=params)[0]
    # print(len(tasks))
    # print(len(state.get_keys()))
    new_key = state.add_onebody_state(state=psi, task=task)
    # print(len(tasks))
    # print(len(state.get_keys())) # there is a problem with the length of tasks
    # TODO: should we make a deep copy of the tasks dict

    # add several states as a dict
    new_tasks = manybody.calc_tasks(intervals, spectrum, occupation, keys=itertools.count(50))
    new_psi = manybody.calc_initial_state(syst, new_tasks, boundaries, params=params)
    old_keys = state.get_keys()
    state.add_distributed_onebody_states(states=new_psi, tasks=new_tasks)
    assert len(state.get_keys()) == len(old_keys) + len(new_tasks)
    # TODO: maybe throw an error if keys already present, or at least with flag

    # add a state with user provided key
    my_new_key = 80
    new_key = state.add_onebody_state(state=psi, task=task, key=my_new_key)
    assert new_key == my_new_key
    # try to add twice with the same key
    raises(ValueError, state.add_onebody_state, state=psi, task=task, key=my_new_key)

    # retrieve a onebody state
    key = 1
    psi_retrieved = state.get_onebody_state(key=key)
    assert isinstance(psi_retrieved, onebody.WaveFunction)
    assert psi_retrieved.energy == tasks[key].energy
    # try to get a state that does not exist
    new_key = state.get_free_key()
    raises(ValueError, state.get_onebody_state, key=new_key)

    # delete a onebody state
    key = 4
    keys = state.get_keys()
    assert key in keys
    state.delete_onebody_state(key=key)
    assert key not in state.get_keys()
    raises(ValueError, state.delete_onebody_state, key)  # TODO: should we raise a key error here?
    assert len(state.get_keys()) == len(keys) - 1

    # evaluate an operator only on a subset of wavefunctions
    state.evolve(time=10)
    density = state.evaluate(density_operator)
    even_keys = [key for key in state.get_keys() if not key % 2]
    odd_keys = [key for key in state.get_keys() if key % 2]

    dens_even = state._calc_expectation_value(density_operator, keys=even_keys)
    dens_odd = state._calc_expectation_value(density_operator, keys=odd_keys)
    dens = state._calc_expectation_value(density_operator)
    assert_array_almost_equal(dens_even + dens_odd, dens)
    assert_array_almost_equal(dens, density)

    # get the shape of the weighting factor
    assert state.weight_shape() == tasks[0].weight.shape

    # check consistency
    assert state._check_consistency()
    del state.tasks[5]  # make state inconsistent
    assert state._check_consistency() is False

    # check wavefunction with the gauss-kronrod quadrature
    for interval in intervals:
        interval.quadrature = 'kronrod'
    tasks = manybody.calc_tasks(intervals, spectrum, occupation)
    boundaries = leads.automatic_boundary(spectrum, tmax)
    psi_init = manybody.calc_initial_state(syst, tasks, boundaries,
                                           params=params)

    state = manybody.WaveFunction(psi_init, tasks)
    state.evolve(time)

    # check return type, we expect an array running over all sites or just a single number
    density = state.evaluate(density_operator)
    sum_density = state.evaluate(sum_density_operator)

    assert isinstance(density, np.ndarray)
    assert density.shape == (2, len(syst.sites))
    assert isinstance(sum_density, np.ndarray)
    assert sum_density.shape == (2,)

    # check that if we add a state with a weight shape mismatch, that it crashes
    # TODO: this should fail, maybe one should give a clear error message and check for that case
    class Task:
        weight = 0.5  # not compatible to kronrod weight array
    task = Task()
    psi = onebody.ScatteringStates(syst, lead=0, energy=2.4, tmax=tmax, params=params)[0]
    new_key = state.add_onebody_state(state=psi, task=task)
    density = state.evaluate(density_operator)


@pytest.mark.integtest
def test_manybody_normalization():

    # integrating over all energies and leads, the density
    # should be 1 on all sites and for all times

    syst = make_system(W=3).finalized()
    density_operator = kwant.operator.Density(syst)

    chemical_potential = 100  # some value larger than any band energy
    tmax = 10
    params = {'v': 0.0001, 't0': 0, 'A': 0.00157, 'sigma': 24}

    occupation = manybody.lead_occupation(chemical_potential)

    spectrum = kwantspectrum.spectra(syst.leads)
    intervals = manybody.calc_intervals(spectrum, occupation)
    tasks = manybody.calc_tasks(intervals, spectrum, occupation)
    boundaries = leads.automatic_boundary(spectrum, tmax)
    psi_init = manybody.calc_initial_state(syst, tasks, boundaries, params=params)
    wf = manybody.WaveFunction(psi_init, tasks)

    for time in [0, 5, 10]:
        wf.evolve(time)
        density = wf.evaluate(density_operator)
        assert_array_almost_equal(density, 1, decimal=2)


@pytest.mark.integtest
def test_simple_manybody_regression():

    # a simple regression test to make sure that the result does not change
    # over the time

    syst = make_system(W=3).finalized()
    density_operator = kwant.operator.Density(syst, sum=True)

    chemical_potential = 5
    tmax = 20
    params = {'v': 0.1, 't0': 0, 'A': 0.157, 'sigma': 24}

    occupation = manybody.lead_occupation(chemical_potential)

    spectrum = kwantspectrum.spectra(syst.leads)
    intervals = manybody.calc_intervals(spectrum, occupation)
    tasks = manybody.calc_tasks(intervals, spectrum, occupation)
    boundaries = leads.automatic_boundary(spectrum, tmax)
    psi_init = manybody.calc_initial_state(syst, tasks, boundaries, params=params)
    wf = manybody.WaveFunction(psi_init, tasks)

    ref_densities = [53.320067790628464, 24.267566625328836, 81.1847296491332]
    for i, time in enumerate([0, 10, 20]):
        wf.evolve(time)
        density = wf.evaluate(density_operator)[1]
        assert_almost_equal(density, ref_densities[i], decimal=3)


# ---- check state
@pytest.mark.integtest  # marker for integration tests
def test_manybody_state():

    syst = make_system().finalized()
    density_operator = kwant.operator.Density(syst)

    chemical_potential = 3
    tmax = 500
    time = 5
    params = {'v': 0.0001, 't0': 0, 'A': 0.00157, 'sigma': 24}

    occupation = manybody.lead_occupation(chemical_potential)

    # reference result with the manybody wavefunction
    spectrum = kwantspectrum.spectra(syst.leads)
    intervals = manybody.calc_intervals(spectrum, occupation)
    tasks = manybody.calc_tasks(intervals, spectrum, occupation)
    boundaries = leads.automatic_boundary(spectrum, tmax)
    psi_init = manybody.calc_initial_state(syst, tasks, boundaries, params=params)
    wf = manybody.WaveFunction(psi_init, tasks)
    wf.evolve(time)
    density_wf = wf.evaluate(density_operator)[1]

    # manybody state
    state = manybody.State(syst, tmax, occupation, params=params, refine=False)

    state.evolve(time)
    assert state.time == time
    # the manybody state has no public params stored, only each onebody state
    # has a public params dictionary
    for _, psi in state.manybody_wavefunction.psi._data.items():
        assert psi.time == time
        assert psi.params == {'v': 0.0001, 't0': 0, 'A': 0.00157, 'sigma': 24}

    # check that result is similar to reference result
    density = state.evaluate(density_operator)
    assert_array_almost_equal(density, density_wf)

    # -- test get_intervals
    intervals = state.get_intervals()
    assert isinstance(intervals, list)
    assert np.array([isinstance(x, manybody.Interval) for x in intervals]).all()

    # -- test estimate_error
    error_per_interval = state.estimate_error(intervals=intervals)
    assert isinstance(error_per_interval, np.ndarray)
    assert len(error_per_interval) == len(intervals)
    assert np.array([isinstance(x, float) for x in error_per_interval]).all()

    err_sum = 0
    for i, interval in enumerate(intervals):
        error = state.estimate_error(intervals=interval)
        assert isinstance(error, float)
        assert error == error_per_interval[i]
        err_sum += state.estimate_error(intervals=interval, full_output=True)
        assert isinstance(err_sum, np.ndarray)

    # the total error is NOT the sum of interval errors (triangle inequality)
    total_error = state.estimate_error()
    assert np.allclose(total_error, np.max(err_sum))

    # -- unknown interval should raise an error
    interval = intervals[0]
    interval.kmin += 0.5 * (interval.kmax - interval.kmin)
    raises(KeyError, state.estimate_error, intervals=interval)

    # -- test refine_intervals
    # -- test time before refining the intervals
    for key in state.manybody_wavefunction.get_keys():
        psi = state.manybody_wavefunction.get_onebody_state(key)
        assert psi.time == state.time
    # -- refine_intervals
    error_before = state.estimate_error()
    state.refine_intervals(atol=1E-3, rtol=1E-3)
    error_after = state.estimate_error()
    assert error_after < error_before
    # -- test time after refining the intervals
    for key in state.manybody_wavefunction.get_keys():
        psi = state.manybody_wavefunction.get_onebody_state(key)
        assert psi.time == state.time

    # add boundstates.
    # emulate the boundstates by some arbitrary onebody states
    # as the boundstates have zero weight, the result must be indentical
    # to the result without boundstates
    def make_psi(energy):
        return onebody.ScatteringStates(syst, lead=0, energy=energy,
                                        tmax=tmax, params=params)[0]
    boundstate_tasks = {'b1': onebody.Task(weight=0, energy=2.1, lead=None, mode=None),
                        'b2': onebody.Task(weight=0, energy=2.3, lead=None, mode=None)}
    boundstate_psi = {key: make_psi(task.energy) for key, task in boundstate_tasks.items()}

    # -- check that adding boundstates do not change error, intervals and observable
    intervals_before = state.get_intervals()
    error_per_interval_before = state.estimate_error(intervals=intervals_before)
    total_error_before = state.estimate_error()
    density_before = state.evaluate(density_operator)
    keys_before = state.manybody_wavefunction.get_keys()

    state._add_boundstates(boundstate_psi, boundstate_tasks)

    intervals_after = state.get_intervals()
    error_per_interval_after = state.estimate_error(intervals=intervals_after)
    total_error_after = state.estimate_error()
    density_after = state.evaluate(density_operator)
    keys_after = state.manybody_wavefunction.get_keys()

    for interval_before, interval_after in zip(intervals_before, intervals_after):
        assert interval_before == interval_after
    assert total_error_before == total_error_after
    assert_array_almost_equal(error_per_interval_before, error_per_interval_after)
    assert_array_almost_equal(density_before, density_after)

    assert keys_before + list(boundstate_tasks.keys()) == keys_after

    # -- check that the state can still be evolved and gives the same result as wavefunction
    time += 5
    state.evolve(time)
    density = state.evaluate(density_operator)

    wf.evolve(time)
    density_wf = state.evaluate(density_operator)

    assert_array_almost_equal(density, density_wf)

    # -- check that refine_intervals still works
    error_before = state.estimate_error()
    state.refine_intervals(atol=1E-4, rtol=1E-4)
    error_after = state.estimate_error()
    assert error_after < error_before

    # -- check to add user given intervals, precalculate
    ModInterval = functools.partial(manybody.Interval, order=3, integration_variable='energy')
    intervals = manybody.calc_intervals(spectrum, occupation, interval_type=ModInterval)
    state = manybody.State(syst, tmax, occupation, params=params, intervals=intervals, refine=False)
    intervals = state.get_intervals()
    for interval in intervals:
        assert interval.order == 3
        assert interval.integration_variable == 'energy'

    # -- check to add user given intervals, new default
    ModInterval = functools.partial(manybody.Interval, order=3, integration_variable='energy')
    state = manybody.State(syst, tmax, occupation, params=params, intervals=ModInterval, refine=False)
    intervals = state.get_intervals()
    for interval in intervals:
        assert interval.order == 3
        assert interval.integration_variable == 'energy'

    # -- check state with default occupation
    occupation_mu_0 = manybody.lead_occupation(chemical_potential=0)
    state_mu_0 = manybody.State(syst, tmax, occupation_mu_0, params=params, refine=False)
    state_default = manybody.State(syst, tmax=tmax, params=params, refine=False)
    assert state_mu_0.occupations[0].energy_range == state_default.occupations[0].energy_range

    # test for missing input
    raises(ValueError, manybody.State, syst=syst, params=params)
    # test for system not finalized
    raises(TypeError, manybody.State, syst=make_system(), tmax=5, params=params)
    # test for mutual exclusive arguments
    raises(ValueError, manybody.State, syst, tmax, boundaries=boundaries, params=params)

    params_with_time = {'v': 0.1, 't0': 0, 'A': 0.00157, 'sigma': 24, 'time': 0}
    with pytest.raises(KeyError) as exc:
        manybody.State(syst, tmax, params=params_with_time)
    error_string = str(exc.value).strip('\"')  # remove etra quotes
    assert error_string == "'params' must not contain time"

    # TODO, check general parameter handling in manybody.State
    # no params passed at all
    # with pytest.raises(TypeError) as exc:
    # manybody.State(syst, tmax)
    # assert str(exc.value) == 'System is missing required arguments: "v"'

    # params_without_sigma = {'v': 0.1, 't0': 0, 'A':0.00157}
    # with pytest.raises(TypeError) as exc:
    #    manybody.State(syst, tmax, params=params_without_sigma)
    # assert str(exc.value) == 'System is missing required arguments: "sigma"'


@pytest.mark.integtest  # marker for integration tests
def test_manybody_state_initial_refine():

    syst = make_system().finalized()

    tmax = 500
    params = {'v': 0.0001, 't0': 0, 'A': 0.00157, 'sigma': 24}

    occupation = manybody.lead_occupation(chemical_potential=3)

    state = manybody.State(syst, tmax, occupation, params=params, refine=False)
    intervals = state.get_intervals()
    assert len(intervals) == 2

    state = manybody.State(syst, tmax, occupation, params=params)
    intervals = state.get_intervals()
    assert len(intervals) == 13


@pytest.mark.integtest  # marker for integration tests
def test_manybody_state_refine_non_kronrod():

    # test that refinement acts only on the gauss-kronrod intervals
    # and that additional other intervals (here gauss-legendre in the example)
    # are not touched by refine_intervals()

    syst = make_system().finalized()

    tmax = 500
    params = {'v': 0.0001, 't0': 0, 'A': 0.00157, 'sigma': 24}

    occupation = manybody.lead_occupation(chemical_potential=3)

    state = manybody.State(syst, tmax, occupation, params=params, refine=False)
    intervals = state.get_intervals()
    assert len(intervals) == 2

    for interval in intervals:
        interval.quadrature = 'gausslegendre'
        state._add_interval(interval)

    intervals = state.get_intervals()
    assert len(intervals) == 4  # 2 + 2 intervals

    state.refine_intervals()
    intervals = state.get_intervals()
    assert len(intervals) == 15  # 13 + 2 intervals


@pytest.mark.integtest
def test_state_estimate_error():

    syst = make_system().finalized()
    density_sum = kwant.operator.Density(syst, sum=True)
    density_operator = kwant.operator.Density(syst)
    current_operator = kwant.operator.Current(syst)

    chemical_potential = 3
    tmax = 500
    params = {'v': 0.0001, 't0': 0, 'A': 0.00157, 'sigma': 24}

    occupation = manybody.lead_occupation(chemical_potential)
    state = manybody.State(syst, tmax, occupation, params=params, refine=False)

    shape_density_sum = state.evaluate(density_sum).shape
    shape_density = state.evaluate(density_operator).shape
    shape_current = state.evaluate(current_operator).shape

    assert shape_density_sum == ()
    assert shape_density_sum != shape_density  # make sure to measure not the same
    assert shape_density_sum != shape_current  # make sure to measure not the same
    assert shape_density != shape_current  # make sure to measure not the same

    intervals = state.get_intervals()

    # --- density is the default operator to measure the error

    # maximum error, summed over all intervals
    errors = state.estimate_error()
    assert errors.shape == ()

    # error over array, summed over all intervals
    errors = state.estimate_error(full_output=True)
    assert errors.shape == shape_density

    # maximum error per interval
    errors = state.estimate_error(intervals=intervals)
    assert errors.shape == (len(intervals),)

    # error over array per interval
    errors = state.estimate_error(intervals=intervals, full_output=True)
    assert errors.shape == (len(intervals),) + shape_density

    # maximum error of a single interval
    errors = state.estimate_error(intervals=intervals[0])
    assert errors.shape == ()

    # error over array of a single interval
    errors = state.estimate_error(intervals=intervals[0], full_output=True)
    assert errors.shape == shape_density

    # --- measure the error with the current operator

    # maximum error, summed over all intervals
    errors = state.estimate_error(error_op=current_operator)
    assert errors.shape == ()

    # error over array, summed over all intervals
    errors = state.estimate_error(error_op=current_operator, full_output=True)
    assert errors.shape == shape_current

    # maximum error per interval
    errors = state.estimate_error(error_op=current_operator, intervals=intervals)
    assert errors.shape == (len(intervals),)

    # error over array per interval
    errors = state.estimate_error(error_op=current_operator, intervals=intervals, full_output=True)
    assert errors.shape == (len(intervals),) + shape_current

    # -- an operator with a scalar output

    # maximum error, summed over all intervals
    errors = state.estimate_error(error_op=density_sum)
    assert errors.shape == ()

    # error over array, summed over all intervals
    errors = state.estimate_error(error_op=density_sum, full_output=True)
    assert errors.shape == shape_density_sum

    # maximum error per interval
    errors = state.estimate_error(error_op=density_sum, intervals=intervals)
    assert errors.shape == (len(intervals),)

    # error over array per interval
    errors = state.estimate_error(error_op=density_sum, intervals=intervals, full_output=True)
    assert errors.shape == (len(intervals),) + shape_density_sum

    # --- set current operator as default from the beginning

    state = manybody.State(syst, tmax, occupation, params=params,
                           error_op=current_operator, refine=False)

    errors = state.estimate_error()
    assert errors.shape == ()

    # error over array, summed over all intervals
    errors = state.estimate_error(full_output=True)
    assert errors.shape == shape_current

    # maximum error per interval
    errors = state.estimate_error(intervals=intervals)
    assert errors.shape == (len(intervals),)

    # error over array per interval
    errors = state.estimate_error(intervals=intervals, full_output=True)
    assert errors.shape == (len(intervals),) + shape_current


@pytest.mark.integtest  # marker for integration tests
def test_manybody_state_estimate_error_non_kronrod():

    syst = make_system().finalized()

    tmax = 500
    params = {'v': 0.0001, 't0': 0, 'A': 0.00157, 'sigma': 24}

    occupation = manybody.lead_occupation(chemical_potential=3)

    state = manybody.State(syst, tmax, occupation, params=params, refine=False)
    intervals = state.get_intervals()
    gauss_interval = intervals[0]
    gauss_interval.quadrature = 'gausslegendre'

    with pytest.raises(ValueError) as exc:
        state.estimate_error(intervals=gauss_interval)
    assert str(exc.value) in "quadpack error estimate works only on Gauss-Kronrod quadrature intervals"


@pytest.mark.integtest  # marker for integration tests
def test_time_name_and_start_default_change_manybody_state():

    N = 3
    lat = kwant.lattice.square(norbs=2)
    I0 = ta.identity(2)
    SZ = ta.array([[1, 0], [0, -1]])
    uniform = kwant.digest.uniform

    tmax = 100
    params = {'salt': '1'}
    occupation = manybody.lead_occupation(chemical_potential=1)

    def make_simple_lead(lat, N):
        I0 = ta.identity(lat.norbs)
        syst = kwant.Builder(kwant.TranslationalSymmetry((-1, 0)))
        syst[(lat(0, j) for j in range(N))] = 4 * I0
        syst[lat.neighbors()] = -1 * I0
        return syst

    def td_onsite_time(site, time, salt):
        static_part = (4 + uniform(site.tag, salt=salt)) * I0
        td_part = SZ * sin(time)
        return static_part + td_part

    def phi_time(time, salt):
        return (1 - cos(time)) * ta.array([1, -1])  # diagonal part

    zeit_offset = 42  # the initial time

    def td_onsite_zeit(site, zeit, salt):
        time = zeit - zeit_offset
        static_part = (4 + uniform(site.tag, salt=salt)) * I0
        td_part = SZ * sin(time)
        return static_part + td_part

    def phi_zeit(zeit, salt):
        time = zeit - zeit_offset
        return (1 - cos(time)) * ta.array([1, -1])  # diagonal part

    # test when adding time-dependent voltage everywhere
    syst = make_td_system_with_leads(lat, N, make_simple_lead, td_onsite_time)
    leads.add_voltage(syst, 0, phi_time)
    leads.add_voltage(syst, 1, phi_time)
    fsyst_time = syst.finalized()

    syst = make_td_system_with_leads(lat, N, make_simple_lead, td_onsite_zeit)
    leads.add_voltage(syst, 0, phi_zeit)
    leads.add_voltage(syst, 1, phi_zeit)
    fsyst_zeit = syst.finalized()

    density_operator_time = kwant.operator.Density(fsyst_time)
    density_operator_zeit = kwant.operator.Density(fsyst_zeit)

    # default
    state_time = manybody.State(fsyst_time, tmax, occupation, params=params, refine=False)

    # default changed
    WaveFunction_zeit = functools.partial(onebody.WaveFunction.from_kwant, time_name='zeit',
                                          time_start=zeit_offset)
    scattering_state_zeit = functools.partial(onebody.ScatteringStates,
                                              wavefunction_type=WaveFunction_zeit)
    state_zeit = manybody.State(fsyst_zeit, tmax, occupation, params=params,
                                scattering_state_type=scattering_state_zeit, refine=False)

    for time in (0., 1., 5.):

        state_time.evolve(time)
        state_zeit.evolve(time + zeit_offset)

        for _, psi in state_time.manybody_wavefunction.psi._data.items():
            assert psi.time == time
            assert psi.time_start == 0
            assert psi.time_name == 'time'

        for _, psi in state_zeit.manybody_wavefunction.psi._data.items():
            assert psi.time == time + zeit_offset
            assert psi.time_start == zeit_offset
            assert psi.time_name == 'zeit'

        # check that result is similar to default result
        density_time = state_time.evaluate(density_operator_time)
        density_zeit = state_time.evaluate(density_operator_zeit)
        assert_array_almost_equal(density_time, density_zeit)

    for t0 in [1 + 1j, 'a', None, -float('inf'), float('inf')]:
        WaveFunction_zeit = functools.partial(onebody.WaveFunction.from_kwant, time_name='zeit',
                                              time_start=t0)
        scattering_state_zeit = functools.partial(onebody.ScatteringStates,
                                                  wavefunction_type=WaveFunction_zeit)
        with pytest.raises(TypeError) as exc:
            manybody.State(fsyst_zeit, tmax, occupation, params=params,
                           scattering_state_type=scattering_state_zeit, refine=False)
        assert str(exc.value) == "time must be a finite real number"


@pytest.mark.integtest
def test_evaluate_integrand():

    syst = make_system().finalized()
    density_operator = kwant.operator.Density(syst)

    tmax = 10
    time = 1  # timestep we like to evaluate the integrand
    params = {'v': 0.0001, 't0': 0, 'A': 0.00157, 'sigma': 24}

    occupation = manybody.lead_occupation(chemical_potential=3)
    state = manybody.State(syst, tmax, occupation, params=params, refine=False)
    state.evolve(time)

    # calculate the integrand on an integral with the state
    interval = manybody.Interval(lead=0, band=0, kmin=0.02, kmax=0.1)
    state._add_interval(interval)
    density_shape = state.evaluate(density_operator).shape
    abcissa, integrand = state.calc_integrand(density_operator, interval)

    # test type and shape
    assert isinstance(abcissa, np.ndarray)
    assert isinstance(integrand, np.ndarray)
    assert abcissa.shape == (2 * interval.order + 1,)
    assert integrand.shape == (2 * interval.order + 1,) + density_shape

    # evaluate the integrand with the separate debug routine on all abcissa point
    # and check if we obtain identical results
    manybody_integrand = manybody.ManybodyIntegrand(syst, interval, density_operator,
                                                    time=time, params=params)
    integ = np.array([manybody_integrand.func(k) for k in abcissa])

    assert manybody_integrand.func(0.05).shape == density_shape
    assert_array_almost_equal(integ, integrand)


def test_calc_energy_cutoffs():

    class Occupation:
        def __init__(self, energy_range):
            self.energy_range = energy_range
            self.multivariate = False

    # ----- just one lead

    occupation = Occupation(energy_range=[(-5, 1), (2, 3)])
    emin, emax = manybody.calc_energy_cutoffs(occupation)

    assert emin == -5
    assert emax == 3

    occupation = Occupation(energy_range=[(None, 1), (2, 3)])
    emin, emax = manybody.calc_energy_cutoffs(occupation)

    assert emin is None
    assert emax == 3

    occupation = Occupation(energy_range=[(-5, 1), (2, None)])
    emin, emax = manybody.calc_energy_cutoffs(occupation)

    assert emin == -5
    assert emax is None

    occupation = Occupation(energy_range=[(None, 1), (2, None)])
    emin, emax = manybody.calc_energy_cutoffs(occupation)

    assert emin is None
    assert emax is None

    # ----- several leads

    occup1 = Occupation(energy_range=[(-5, 1)])
    occup2 = Occupation(energy_range=[(-4, 1.5), (2, 3)])
    emin, emax = manybody.calc_energy_cutoffs([occup1, occup2])

    assert emin == -5
    assert emax == 3

    occup1 = Occupation(energy_range=[(-5, None)])
    occup2 = Occupation(energy_range=[(None, 1.5), (2, 3)])
    emin, emax = manybody.calc_energy_cutoffs([occup1, occup2])

    assert emin is None
    assert emax is None

    # ----- several leads, one lead not occupied

    occup1 = Occupation(energy_range=[(-5, 1)])
    occup2 = Occupation(energy_range=[(-4, 1.5), (2, 3)])

    emin, emax = manybody.calc_energy_cutoffs([occup1, occup2, None])

    assert emin == -5
    assert emax == 3

    # ----- no lead is occupied
    emin, emax = manybody.calc_energy_cutoffs(None)

    assert emin is None
    assert emax is None


def test_boundstates_present():

    syst = make_system_with_boundstate(V=0).finalized()
    assert not manybody.boundstates_present(syst)

    syst = make_system_with_boundstate(V=-5).finalized()
    assert manybody.boundstates_present(syst)


@pytest.mark.integtest  # marker for integration tests
def test_add_boundstate():

    syst = make_system_with_boundstate(V=-5).finalized()

    hamiltonian = syst.hamiltonian_submatrix()
    eigenvalues, eigenvectors = np.linalg.eigh(hamiltonian)

    bs_energy = eigenvalues[0]
    psi_bs = eigenvectors[:, 0]

    occupations = manybody.lead_occupation(chemical_potential=4)
    density_operator = kwant.operator.Density(syst)

    state = manybody.State(syst, tmax=1, occupations=occupations)
    state.add_boundstate(psi_bs, bs_energy)
    density = state.evaluate(density_operator, root=None)
    assert_array_almost_equal(density, 1, decimal=4)


@pytest.mark.integtest  # marker for integration tests
def test_single_spin_rotating_basis():
    # test against analytica result for a single spin:
    # Son-Hsien Chen, Ching-Ray Chang, John Q. Xiao, and Branislav K. Nikolić, 
    # Spin and charge pumping in magnetic tunnel junctions with precessing magnetization: 
    # A nonequilibrium Green function approach, Phys. Rev. B 79, 054424 (2009).
    # $I^z = \frac{2 e}{\hbar} J^{S_z}$, and $J^{S_z}$ is the expression from Eq. (13):

    def sigma(ef):
        sqrtarg = ef**2 - 4
        if sqrtarg < 0:
            sqrt = 1j * np.sqrt(-sqrtarg)
        else:
             sqrt = np.sqrt(sqrtarg)       
        return ef - sqrt / 2

    def rfunc(sigma, J):
        return (4 * sigma**2 - J**2) * sigma**2 / 4 - 2 * sigma**2 + 1

    def jz(J, omega, theta, ef):  # I^z
        sig = sigma(ef)
        r = rfunc(sig, J)
        return omega * np.sin(theta)**2 * J**2 * (sig.imag)**2 / (np.pi * np.abs(r)**2) * (1 + np.abs(sig)**2)

    # parameters
    a = 1 # lattice constant
    w = 0.01 # Frequency of oscillation in units of gamma/hbar
    J = 0.1 # Coupling between electrons and magnetic moment in units of gamma
    theta = np.pi / 4 # Cone angle of the magnetic moments
    Ef = 0.0 # Fermi energy in units of gamma

    syst, lat = make_single_spin_rotating_basis(a, J, w, theta)
    syst = syst.finalized()

    occupations = manybody.lead_occupation(chemical_potential=[Ef-w/2, Ef+w/2])

    # operator to measure the current I^z from the system to the left lead (from site 0 to site -1)
    sigma_z = ta.array([[1,0], [0, -1]])
    iz_operator = kwant.operator.Current(syst, sigma_z, where=[(lat(-1, 0), lat(0, 0))])

    tmax = 1  # no time evolution needed
    state = manybody.State(syst, tmax, occupations)

    current_z = state.evaluate(iz_operator)
    current_ref = jz(J, w, theta, Ef)

    assert_almost_equal(current_z, current_ref, decimal=8)



# -*- coding: utf-8 -*-
# Copyright 2016 tkwant authors.
#
# This file is part of tkwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# https://tkwant.kwant-project.org/doc/stable/pre/license.html.
# A list of tkwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# https://tkwant.kwant-project.org/doc/stable/pre/authors.html.
"""Common utilities for testing"""

import numpy as np
import tinyarray as ta
import kwant


def make_square(N, lat=None, sym=kwant.builder.NoSymmetry(), norbs=1, functional=False, mat=np.eye):
    if not lat:
        lat = kwant.lattice.square(norbs=norbs)
    syst = kwant.Builder(sym)
    onsite = (lambda i, t, arg: 4 * mat(norbs)) if functional else 4 * mat(norbs)
    hop = (lambda i, j, t, arg: -1 * mat(norbs)) if functional else -1 * mat(norbs)
    syst[(lat(i, j) for i in range(N) for j in range(N))] = onsite
    syst[lat.neighbors()] = hop
    return lat, syst


def make_chain(N, lat=None, sym=kwant.builder.NoSymmetry(), norbs=1, functional=False, mat=np.eye):
    if not lat:
        lat = kwant.lattice.chain(norbs=norbs)
    syst = kwant.Builder(sym)
    onsite = (lambda i, t, arg: 2 * mat(norbs)) if functional else 2 * mat(norbs)
    hop = (lambda i, j, t, arg: -1 * mat(norbs)) if functional else -1 * mat(norbs)
    syst[(lat(i) for i in range(N))] = onsite
    syst[lat.neighbors()] = hop
    return lat, syst


def make_square_with_leads(N, norbs=1, functional=False, mat=np.eye):
    lat, syst = make_square(N, norbs=norbs, mat=mat)
    _, lead = make_square(N, lat, kwant.TranslationalSymmetry([-1, 0]),
                          norbs=norbs, functional=functional, mat=mat)
    syst.attach_lead(lead)
    syst.attach_lead(lead.reversed())
    return lat, syst


def make_chain_with_leads(N, norbs=1, functional=False, mat=np.eye):
    lat, syst = make_chain(N, norbs=norbs, mat=mat)
    _, lead = make_chain(N, lat, kwant.TranslationalSymmetry([-1]),
                         norbs=norbs, functional=functional, mat=mat)
    syst.attach_lead(lead)
    syst.attach_lead(lead.reversed())
    return lat, syst


def make_simple_lead(lat):
    syst = kwant.Builder(kwant.TranslationalSymmetry((-1, 0)))
    syst[(lat(0, j) for j in range(3))] = 4
    syst[lat.neighbors()] = -1 + 1j
    return syst


def make_complex_lead(lat):
    syst = kwant.Builder(kwant.TranslationalSymmetry((-1, 0)))
    syst[(lat(0, j) for j in range(3))] = 4
    syst[kwant.HoppingKind((0, 1), lat)] = -1
    syst[(lat(0, 0), lat(1, 0))] = -1j
    syst[(lat(0, 1), lat(1, 0))] = -1
    syst[(lat(0, 2), lat(1, 2))] = -1 + 1j
    return syst


def make_system_with_leads(lat, lead_maker):
    syst = kwant.Builder()
    syst[(lat(i, j) for i in range(3) for j in range(3))] = 4
    syst[lat.neighbors()] = -1
    syst.attach_lead(lead_maker(lat))
    syst.attach_lead(lead_maker(lat).reversed())
    return syst


def check_boundary_hamiltonian(H, cell_norbs, iface_norbs,
                               cell_hamiltonian, inter_cell_hopping):
    this_cell = ta.array([0, cell_norbs])
    this_iface = ta.array([0, iface_norbs])
    H = H.todense()
    ncells = H.shape[0] // cell_norbs
    for i in range(ncells):
        H_cell = cell_hamiltonian(i)
        assert np.allclose(H[slice(*this_cell), slice(*this_cell)], H_cell,
                           atol=1e-16, rtol=1e-16)
        # can only check forwards/backwards if we are not on the last/first cell
        if i < ncells - 1:
            V_cell = inter_cell_hopping(i)
            next_cell = this_cell + cell_norbs
            assert np.allclose(H[slice(*this_iface), slice(*next_cell)],
                               V_cell.conjugate().transpose(), atol=1e-16,
                               rtol=1e-16)
            assert np.allclose(H[slice(*next_cell), slice(*this_iface)],
                               V_cell, atol=1e-16, rtol=1e-16)
        this_cell += cell_norbs
        this_iface += cell_norbs



from math import sin, pi
import matplotlib.pyplot as plt

import tkwant
import kwant


def am_master():
    """Return true for the MPI master rank"""
    return tkwant.mpi.get_communicator().rank == 0


def make_fabry_perot_system():

    # Define an empty tight-binding system on a square lattice.
    lat = kwant.lattice.square(norbs=1)
    syst = kwant.Builder()

    # Central scattering region.
    syst[(lat(x, 0) for x in range(80))] = 0
    syst[lat.neighbors()] = -1
    # Backgate potential.
    syst[(lat(x, 0) for x in range(5, 75))] = -0.0956
    # Barrier potential.
    syst[[lat(4, 0), lat(75, 0)]] = 5.19615

    # Attach lead on the left- and on the right-hand side.
    sym = kwant.TranslationalSymmetry((-1, 0))
    lead = kwant.Builder(sym)
    lead[(lat(0, 0))] = 0
    lead[lat.neighbors()] = -1
    syst.attach_lead(lead)
    syst.attach_lead(lead.reversed())

    return syst, lat


# Phase from the time integrated voltage V(t).
def phi(time):
    vb, tau = 0.6, 30.
    if time > tau:
        return vb * (time - tau / 2.)
    return vb / 2. * (time - tau / pi * sin(pi * time / tau))


def main():

    times = range(220)

    # Make the system and add voltage V(t) to the left lead (index 0).
    syst, lat = make_fabry_perot_system()
    tkwant.leads.add_voltage(syst, 0, phi)
    syst = syst.finalized()

    # Define an operator to measure the current from site 77 to site 78
    # on the right side after the barrier.
    hoppings = [(lat(78, 0), lat(77, 0))]
    current_operator = kwant.operator.Current(syst, where=hoppings)

    # Set occupation T = 0 and mu = -1 for both leads.
    occup = tkwant.manybody.lead_occupation(chemical_potential=-1)

    # Initialize the time-dependent manybody state. Here we use a lower
    # accuracy for the initial adaptive refinement (at time=0) to speed up the calculation.
    # For a realistic simulation, set rtol and atol to smaller values.
    state = tkwant.manybody.State(syst, tmax=max(times), occupations=occup, refine=False)
    state.refine_intervals(rtol=0.3, atol=0.3)

    # Loop over timesteps and evaluate the current.
    currents = []
    for time in times:
        state.evolve(time)
        # For a realistic simulation, one must set rtol and atol to smaller values,
        # or remove both arguments completely, to converge the manybody integral.
        state.refine_intervals(rtol=0.3, atol=0.3)
        current = state.evaluate(current_operator)
        currents.append(current)

    # Plot the normalized current vs. time.
    if am_master():
        plt.plot(times, currents / currents[-1])
        plt.xlabel(r'time $t$')
        plt.ylabel(r'current $I$')
        plt.show()

if __name__ == '__main__':
    main()



# content of conftest.py

import pytest


def pytest_addoption(parser):
    parser.addoption(
        "--integtest", action="store_true", help="run integration tests"
    )
    parser.addoption(
        "--mpitest", action="store_true", help="run mpi tests"
    )


def pytest_runtest_setup(item):
    if 'integtest' in item.keywords and not item.config.getoption("--integtest"):
        pytest.skip("need --integtest option to run this test")
    if 'mpitest' in item.keywords and not item.config.getoption("--mpitest"):
        pytest.skip("need --mpitest option to run this test")



# Script 2/2 to produce the figure in Tkwant's tutorial
# "Self-consistent Tkwant: a generic solver for time-dependent mean field calculations"
# in the section on Microscopic Bogoliubov-deGennes equations.
#
# Step 1: run the script bdg_junction_run_computation.py (preferably on a cluster using MPI parallelization)
# Step 2: run this script to plot the results
#
import matplotlib.pyplot as plt
import numpy as np
import pickle
import pylab

fac = 6
fontsize = 45

pylab.rcParams['figure.figsize'] = (fac * np.sqrt(2), fac)
plt.rcParams.update({'font.size': fontsize})
plt.rc('text', usetex=True)
plt.rc('font', family='serif')

data = pickle.load(open('self_consistent_bdg_data.npy', 'rb'))

times = data['times']
vj = data['v0t'] - data['v']
delta = data['delta']

times *= delta
vj *= 1 / delta

plt.plot(times, vj)

plt.xlabel(r"$t\ \mathrm{[\hbar/\Delta]}$", fontsize=fontsize)
plt.ylabel(r"$V_J \ \mathrm{[\Delta/e]}$", fontsize=fontsize)

plt.savefig("bdg_junction_vj_vs_time.png", bbox_inches='tight')

plt.show()



# Copyright 2016-2023 tkwant authors.
#
# This file is part of tkwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# https://tkwant.kwant-project.org/doc/stable/pre/license.html.
# A list of tkwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# https://tkwant.kwant-project.org/doc/stable/pre/authors.html.
"""Utility routines."""


import subprocess
import os
import inspect
import collections
import numpy as np

__all__ = ['version', 'TkwantDeprecationWarning', 'is_type', 'is_type_array',
           'is_not_empty', 'is_zero', 'time_start', 'time_name', 'memoize']

# tkwant's default initial time and time argument name
time_start = 0
time_name = 'time'

TKWANT_VERSION = "1.1.0"

package_root = os.path.dirname(os.path.realpath(__file__))
distr_root = os.path.dirname(package_root)


class TkwantDeprecationWarning(Warning):
    """Class of warnings about a deprecated feature of tkwant.

    DeprecationWarning has been made invisible by default in Python 2.7 in order
    to not confuse non-developer users with warnings that are not relevant to
    them.  In the case of tkwant, by far most users are developers, so we feel
    that a TkwantDeprecationWarning that is visible by default is useful.
    """
    pass


def ensure_isinstance(obj, typ, msg=None):
    if isinstance(obj, typ):
        return
    if msg is None:
        msg = "Expecting an instance of {}.".format(typ.__name__)
    raise TypeError(msg)

# type checking and small helper routines


def is_type(variable, generic_type, require_finite=False):
    """Return true if type(variable) matches the generic type."""
    # TODO: require finite for all types
    if generic_type == 'integer':
        return np.issubdtype(type(variable), np.integer)
    if generic_type == 'number':
        return np.issubdtype(type(variable), np.number)
    if generic_type == 'real_number':
        _type = type(variable)
        is_number = np.issubdtype(_type, np.number)
        is_complex = np.issubdtype(_type, np.complexfloating)
        if is_number and require_finite:
            is_finite = np.isfinite(variable)
            return is_number and is_finite and not is_complex
        return is_number and not is_complex
    if generic_type == 'complex':
        _type = type(variable)
        is_number = np.issubdtype(_type, np.number)
        is_complex = np.issubdtype(_type, np.complexfloating)
        return is_number and is_complex
    raise NotImplementedError('generic_type= {} not implemented'
                              .format(generic_type))


def is_type_array(array, generic_type):
    """Return true everywhere where type(array) matches the generic type"""
    array = np.array(array)
    return np.array([is_type(x, generic_type) for x in array.flatten()])


def is_zero(x, tol=1E-14):
    """Return true if |x| < tol."""
    return np.abs(x) < tol


def is_not_empty(obj):
    """Check if obj exists."""
    try:
        return obj.any() or len(obj) > 0
    except AttributeError:
        if obj:
            return True
        return False


def get_default_function_argument(func, kwarg):
    """Return the default value of a function argument.

    Parameters
    ----------
    func : callable
        Function to inspect.
    kwarg : str
        Function argument for which we like to obtain the default value.

    Returns
    -------
    default : obj
        Default argument of function `func`.
        If func(kwarg=42), the returned value would be 42.

    Notes
    -----
    A TypeError is raised, if the function `func` has no `kwarg` argument
    with a default value.
    """
    P = inspect.Parameter
    pars = inspect.signature(func).parameters  # an *ordered mapping*
    for k, v in pars.items():
        if k == kwarg and v.kind in (P.POSITIONAL_OR_KEYWORD, P.KEYWORD_ONLY):
            if v.default is not inspect._empty:
                return v.default
    try:
        func_name = func.__name__
    except AttributeError:
        func_name = '<unknown func>'   # for functools.partial and friends
    raise ValueError("Function argument {} not present in {}".
                     format(kwarg, func_name))


# memoize from kwant
def _hashable(obj):
    return isinstance(obj, collections.abc.Hashable)


def memoize(f):
    """Decorator to memoize a function that works even with unhashable args.

    This decorator will even work with functions whose args are not hashable.
    The cache key is made up by the hashable arguments and the ids of the
    non-hashable args.  It is up to the user to make sure that non-hashable
    args do not change during the lifetime of the decorator.

    This decorator will keep reevaluating functions that return None.
    """
    def lookup(*args):
        key = tuple(arg if _hashable(arg) else id(arg) for arg in args)
        result = cache.get(key)
        if result is None:
            cache[key] = result = f(*args)
        return result
    cache = {}
    return lookup



# Script 3/3 to produce the figure in Tkwant's tutorial
# "Self-consistent Tkwant: a generic solver for time-dependent mean field calculations"
# in the section on Luttinger liquids.
#
# Step 1: run the two simulation (plasmon_u_0_computation.py, plasmon_u_10_computation.py)
#         preferably on a cluster using MPI
# Step 2: plot the data (plasmon_plot_results.py)
#
import matplotlib.pyplot as plt
import numpy as np
import pickle
import pylab


fac = 6
fontsize = 45


pylab.rcParams['figure.figsize'] = (fac * np.sqrt(2), fac)
plt.rcParams.update({'font.size': fontsize})
plt.rc('text', usetex=True)
plt.rc('font', family='serif')


# ----no interactions
data = pickle.load(open('plasmon_u_0_data.npy', 'rb'))

t = np.array(data['times'])
x = np.array(data['sites'])
z = np.array(data['densities'])

for i, (time, density) in enumerate(zip(t, z)):
    plt.plot(x, 280000 * density + time, '--', color='black')

# ---- with interactions
data = pickle.load(open('plasmon_u_10_data.npy', 'rb'))

t = np.array(data['times'])
x = np.array(data['sites'])
z = np.array(data['densities'])

for i, (time, density) in enumerate(zip(t, z)):
    plt.plot(x, 800000 * density + time, color='black')

v0 = np.sqrt(3)
g = 10
v = v0 * np.sqrt(1 + g / (np.pi * v0))

t0 = 350

plt.ylabel(r'time $t$', fontsize=fontsize)
plt.xlabel(r'site $i$', fontsize=fontsize)

xx = x[3000:6000]
plt.plot([0, max(xx)], [t0, t0 + max(xx)/(v0)], '--b')
plt.plot([0, max(xx)], [t0, t0 + max(xx)/(v)], 'b')

axes = plt.gca()
axes.set_xlim([0, 3000])
axes.set_ylim([500, 1300])
axes.set_xticks([0, 1000, 2000, 3000])
axes.set_yticks([500, 700, 900, 1100])
plt.text(1350, 1320, r'$U=0$')
plt.text(2450, 1320, r'$U=10$')

plt.savefig("plasmon_propagation.png", bbox_inches='tight')

plt.show()



# -*- coding: utf-8 -*-
# cython: embedsignature=True
#
# Copyright 2016-2022 tkwant authors.
#
# This file is part of tkwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# https://tkwant.kwant-project.org/doc/stable/pre/license.html.
# A list of tkwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# https://tkwant.kwant-project.org/doc/stable/pre/authors.html.
"""Calculating the right-hand side of the time-dependent Schrödinger equation.

This module is mainly for internal use.

The classes in this module are used for calculating the following expression:

    ..math:: -i(H_0 ψ + W(t) ψ)

where :math:`H_0` is the time-independent part of the Hamiltonian,
:math:`W(t)` is the time-dependent part, and :math:`ψ` is some vector
defined over the Hilbert space of the problem. We recognise this expression
as the right-hand side of the time-dependent Schrödinger equation (TDSE),
where the left-hand side is :math:`∂ψ/∂t`.

In practice we will want to handle many different cases, including solving the
TDSE on infinite domains, or starting the evolution from an eigenstate.  All
the forms of TDSE that we will need to handle these cases are encompassed by
the above formula. This module is only concerned with calculating expressions
of the above form, not performing the mapping from the various problems. For
more details on the mapping, see `tkwant/onebody/solver.py`.
"""

__all__ = ['Kernel', 'Scipy', 'Simple',
           'default', 'PerturbationExtractor',
           'PerturbationInterpolator']

import numpy as np
import cython
from .. import _common, system, _logging
from scipy.sparse._sparsetools import coo_matvec, csr_matvec
import scipy.sparse as sp


# set module logger
logger = _logging.make_logger(name=__name__)

########## Abstract Base Classes for kernels

cdef class Kernel:
    """ABC for right-hand side of the time-dependent Schrödinger equation.

    Attributes
    ----------
    size : int
        The size of the time-independent part of the Hamiltonian. This
        also sets the size of the solution vector.
    """

    def __init__(self, H0, W, complex[:] psi_st=None, complex[:] work=None):
        raise NotImplementedError()

    def rhs(self, energy=None):
        """Evaluate the RHS of the TDSE and store the result in `dpsidt`."""
        raise NotImplementedError()


### C-level interface to Kernels (needed for C-implemented solvers)




########## Concrete kernel implementations

class Scipy(Kernel):
    """Evaluate the RHS of the Schrödinger equation using scipy sparse.

    Parameters
    ----------
    H0 : `scipy.sparse.base`
        Hamiltonian at ``t=0`` (including any boundary conditions).
    W : callable
        Time-dependent part of the Hamiltonian. Typically the object returned
        by `tkwant.system.extract_perturbation`.
    psi_st : array of complex, optional
        The wavefunction of the initial eigenstate defined over the central
        system (if starting in an initial eigenstate).
    work : `~numpy.ndarray` of complex, optional
        Workarray of size ``H0.shape[0]`` for performance and memory optimization.

    Attributes
    ----------
    H0 : `scipy.sparse.csr_matrix`
        Hamiltonian at ``t=0`` (including any boundary conditions).
    W : callable
        Time-dependent part of the Hamiltonian. Typically the object returned
        by `tkwant.system.extract_perturbation`.
    psi_st : array of complex or `None`
        The wavefunction of the initial eigenstate defined over the central
        system (if starting in an initial eigenstate).
    """

    def __init__(self, H0, W, psi_st=None, work=None):

        if not isinstance(H0, sp.csr.csr_matrix):
            self.H0 = H0.tocsr()  # self.H0 is no reference any more in that case
        else:
            self.H0 = H0

        self.W = W
        self.psi_st = psi_st
        # The size of the time-independent part of the Hamiltonian. This
        # also sets the size of the solution vector.
        self.size = self.H0.shape[0]

        if W is not None:
            self.center_size = self.W.size
        else:
            self.center_size = self.size

        if work is None:
            self._tmp = np.zeros((self.size,), dtype=complex)
        else:
            if work.size < self.center_size:
                raise ValueError('work array size={} must be at least '
                     'the size of the central system={}'
                     .format(work.size, self.center_size))
            self._tmp = work

    def rhs(self, energy=None):

        if energy is None:
            H0 = self.H0
        else:
            H0 = self.H0 - energy * sp.eye(self.size)

        # The approach with a separated class ``_CalcRHS`` to calculate the r.h.s.
        # seems more complicated, but it allows to store the static Hamiltonian
        # as a reference in class ``Scipy(Kernel)`` in ``self.H0``
        # (for manybody calculations there is onle reference per rank).
        # Moreover, the calculation ``H = H0 - E`` is done inplace,
        # and only once when the Scipy(Kernel).rhs() method is called.
        # It is important to precalculate H0 - E at this stage
        # (typically when ``WaveFunction.evolve()`` is called) and not
        # to do the inplace calculation in ``_CalcRHS.__call__()``,
        # as the later method is called very often (by the DGL solver).
        # The implace calculation generates only a temporary object ``H``
        # which not stored in the ``onebody.WaveFunction``, such
        # that it is cleaned by the garbage collector.
        return _CalcRHS(H0, self.W, self.psi_st, self.size, self.center_size, self._tmp)

    def set_W(self, W):
        '''Set the time dependent perturbation W(t)'''
        self.W = W


class _CalcRHS:
    """Calculate the right-hand-side of the time-dependent Schrödinger equation."""
    def __init__(self, H0, W, psi_st, size, center_size, tmp):

        self.H0 = H0
        self.W = W
        self.psi_st = psi_st
        self.size = size
        self.center_size = center_size
        self._tmp = tmp

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def __call__(self, const complex[:] psi, complex[:] dpsidt, const double time):

        # H0 @ psi -> tmp
        cdef complex[:] tmp = self._tmp
        tmp[:] = 0    
        csr_matvec(
            self.H0.shape[0], self.H0.shape[1], self.H0.indptr,
            self.H0.indices, self.H0.data, psi, tmp)

        # W(t) @ (psi + psi_st) + tmp -> tmp
        psi_centre = np.asarray(psi[:self.center_size])
        if self.psi_st is not None:
            psi_centre = psi_centre + self.psi_st
        try:
            self.W.apply(time, psi_centre[:self.center_size], out=tmp[:self.center_size])
        except Exception as exept:
            if self.W is None:
                pass
            else:
                raise exept

        # dpsidt = -1j * tmp
        cdef int i
        for i in range(self.size):
            dpsidt[i] = -1j * tmp[i]


@cython.boundscheck(False)
@cython.wraparound(False)
cdef(double, double, double, double, double) update_times(double t, double t0, double dt):
    '''Return new update times for the Hamiltonian matrix.
    '''
    assert dt > 0
    cdef int quotient
    quotient = int(round((t - t0) / dt, 10))  # integer division // has problems with the numerics
    t0 = t0 + quotient * dt
    # adding dt up we get inaccurate routings
    cdef double t1 = t0 + dt
    cdef double t2 = t0 + 2 * dt
    cdef double t3 = t0 + 3 * dt
    cdef double t4 = t0 + 4 * dt
    return t0, t1, t2, t3, t4


@cython.boundscheck(False)
@cython.wraparound(False)
cdef(double, double, double, double) _get_polynome_coeffs(double t, double t0, double t1, double dt):

    cdef double rdx = 1 / dt
    cdef double fac = dt * dt / 6.

    cdef double a = (t1 - t) * rdx
    cdef double b = (t - t0) * rdx
    cdef double c = (a * a * a - a) * fac
    cdef double d = (b * b * b - b) * fac
    return a, b, c, d


def _herm_conj(value):
    if hasattr(value, 'conjugate'):
        value = value.conjugate()
    if hasattr(value, 'transpose'):
        value = value.transpose()
    return value


_errormsg = ('{0} vector has length {1}, but the system '
             'contains {2} orbitals')


cdef class Interpolation:

    cdef:
        double dt
        double t0, t1, t2, t3, t4
        double atol, rtol
        double fac, facmin, facmax
        double dt_min
        object Ht0, Ht1, Ht2, Ht3, Ht4
        object d2Ht0, d2Ht1, d2Ht2
        object func
        unsigned nnz

    def __init__(self, func, t0, dt_init=0.001, atol=1E-10, rtol=1E-10, dt_min=1E-6, fac=0.9, facmin=0.1, facmax=3):

        # The spline interpolation in this routine is done "on the fly".
        # Quite generally, for a cubic spline, the cubic function f(t) is defined
        # on two neighboring intervals [t0, t1] and [t1, t2],
        # t1 - t0 = t2 - t1 = dt. One needs to know
        # f(t) on t0, t1, and t2, and f''(t) on the endpoints t0, t2.
        # As the Schrödinger equation which calls the interpolant propagates only
        # foreward in time, we do a stepwise interpolation and assume, at least
        # for the optimization, that the time argument monontonically increases
        # when the interpolation function is called. The idea is to reuse the
        # evaluated function and derivative values, and don't calculate them twice.
        # Therefore, one always need to keep 5. timesteps in memory, t0, t1 .. t4.
        # Interpolation is always done between t0, t1 and t2.
        # The points t3 and t4 are used to estimate the 2. derivative on t2
        # using a central finite difference scheme with 5 points. The second
        # derivative on point t0 is known from the step before.
        # Only when this routine is initialized, we have to estimate f''(t0).
        # We use a foreward derivative in this case.

        assert dt_init > 0
        assert dt_min > 0
        assert atol >= 0
        assert rtol >= 0

        self.dt = dt_init
        self.func = func

        self.atol = atol
        self.rtol = rtol

        self.fac = fac
        self.facmin = facmin
        self.facmax = facmax

        self.dt_min = dt_min

        self.t0 = t0
        self.t1 = self.t0 + self.dt
        self.t2 = self.t1 + self.dt
        self.t3 = self.t2 + self.dt
        self.t4 = self.t3 + self.dt

        self.Ht0 = self.func(self.t0)
        self.Ht1 = self.func(self.t1)
        self.Ht2 = self.func(self.t2)
        self.Ht3 = self.func(self.t3)
        self.Ht4 = self.func(self.t4)

        self.nnz = self.Ht0.shape[0]

        Ht02 = self.func(self.t0 + 0.5 * self.dt)
        Ht12 = self.func(self.t0 + 1.5 * self.dt)
        Ht22 = self.func(self.t0 + 2.5 * self.dt)

        rdt2 = 1 / (self.dt * self.dt)
        rdt2h = 4 * rdt2

        # w''(t0), forward finite difference scheme (4. order)
        # on the half interval dt / 2
        cl0 = 15 / 4 * rdt2h
        cl1 = -77 / 6 * rdt2h
        cl2 = 107 / 6 * rdt2h
        cl3 = -13 * rdt2h
        cl4 = 61 / 12 * rdt2h
        cl5 = -5 / 6 * rdt2h
        self.d2Ht0 = np.zeros((self.nnz,), dtype=complex)
        for i in range(self.nnz):
            self.d2Ht0[i] = cl0 * self.Ht0[i] + cl1 * Ht02[i] + cl2 * self.Ht1[i] \
                + cl3 * Ht12[i] + cl4 * self.Ht2[i] + cl5 * Ht22[i]

        # w''(t2), central finite difference scheme (4. order)
        cc2 = - rdt2 / 12
        cc1 = rdt2 * 4 / 3
        cc0 = - 2.5 * rdt2
        self.d2Ht2 = np.zeros((self.nnz,), dtype=complex)
        for i in range(self.nnz):
            self.d2Ht2[i] = cc2 * (self.Ht0[i] + self.Ht4[i]) \
                + cc1 * (self.Ht1[i] + self.Ht3[i]) \
                + cc0 * self.Ht2[i]

        # w''(t1), spline formula
        self.d2Ht1 = np.zeros((self.nnz,), dtype=complex)
        for i in range(self.nnz):
            self.d2Ht1[i] = 1.5 * rdt2 * (self.Ht0[i] + self.Ht2[i] - 2 * self.Ht1[i]) \
                - 0.25 * (self.d2Ht0[i] + self.d2Ht2[i])

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def eval(self, double time, complex[:] out):

        if self.nnz == 0: return

        cdef double c0, c1, c2, rdt2
        cdef double a, b, c, d
        cdef unsigned i

        # update matrices
        time_larger = time >= self.t2
        time_smaller = time < self.t0

        if time_larger or time_smaller:

            if time_larger:
                error = self._estimate_error(self.t1 + 0.5 * self.dt)
            else:
                error = self._estimate_error(self.t0 + 0.5 * self.dt)

            self.dt = self._estimate_stepsize(self.dt, error)

            t0_new, t1_new, t2_new, t3_new, t4_new = update_times(time, self.t0, self.dt)
            rdt2 = 1 / (self.dt * self.dt)
            cc2 = - rdt2 / 12
            cc1 = rdt2 * 4 / 3
            cc0 = - 2.5 * rdt2

            Htm2 = self.func(t0_new - 2 * self.dt)
            Htm1 = self.func(t0_new - self.dt)
            self.func(t0_new, self.Ht0)
            self.func(t1_new, self.Ht1)
            self.func(t2_new, self.Ht2)
            self.func(t3_new, self.Ht3)
            self.func(t4_new, self.Ht4)

            for i in range(self.nnz):
                # w''(t) on t0 and t2, central finite difference scheme
                self.d2Ht0[i] = cc2 * (Htm2[i] + self.Ht2[i]) \
                    + cc1 * (Htm1[i] + self.Ht1[i]) \
                    + cc0 * self.Ht0[i]
                self.d2Ht2[i] = cc2 * (self.Ht0[i] + self.Ht4[i]) \
                    + cc1 * (self.Ht1[i] + self.Ht3[i]) \
                    + cc0 * self.Ht2[i]
                # w''(t) on t1, spline formula
                self.d2Ht1[i] = 1.5 * rdt2 * (self.Ht0[i] + self.Ht2[i] - 2 * self.Ht1[i])\
                    - 0.25 * (self.d2Ht0[i] + self.d2Ht2[i])

            self.t0, self.t1, self.t2, self.t3, self.t4 = t0_new, t1_new, t2_new, t3_new, t4_new

        # construct the spline, either on [t0, t1] or on [t1, t2] interval.
        if time < self.t1:
            a, b, c, d = _get_polynome_coeffs(time, self.t0, self.t1, self.dt)
            for i in range(self.nnz):
                out[i] = a * self.Ht0[i] + b * self.Ht1[i] + c * self.d2Ht0[i] + d * self.d2Ht1[i]
        else:
            a, b, c, d = _get_polynome_coeffs(time, self.t1, self.t2, self.dt)
            for i in range(self.nnz):
                out[i] = a * self.Ht1[i] + b * self.Ht2[i] + c * self.d2Ht1[i] + d * self.d2Ht2[i]

    def _estimate_error(self, double time):
        fx = self.func(time)
        fapp = np.zeros((self.nnz,), dtype=complex)
        self.eval(time, fapp)
        scale = self.atol + self.rtol * np.abs(fx)
        return np.max(np.abs(fx - fapp) / scale)

    def _estimate_stepsize(self, double dt, double error):
        # Eq. (4.13), page 168 in book:
        # E. Hairer, S. Nørsett, and G. Wanner, Solving Ordinary Differential Equations,
        # I: Nonstiff Problems, Second  revised edition (Springer, Berlin, Heidelberg, 2008).
        assert dt > 0
        assert error >= 0
        a = - 1 / 4  # cubic order p = 3
        dt_new = dt * min(self.facmax, max(self.facmin, self.fac * error**a))
        if dt_new < self.dt_min:
            logger.warning('minimal W(t) stepsize dt_min={} reached.'
                           .format(self.dt_min))
            dt_new = self.dt_min
        return dt_new



# TODO: this only works for finalized builders so far -- how to extend
#       this to something more general?
# TODO: change this routine when kwant supports vectorized systems
cdef class PerturbationExtractor:
    """Extract the time-dependent perturbation to the Hamiltonian.

    The total Hamiltonian can be split into two parts: the stationary
    part and the time-dependent perturbation, i.e.  H(t) = H_0 + W(t).
    This routine extracts the W(t) part from the Kwant system that
    specifies the H(t).

    Parameters
    ----------
    syst : `kwant.builder.FiniteSystem`
        The system from which to extract the time-dependent perturbation.
    time_name : str
        The name of the time argument. Only sites with Hamiltonian value
        functions with the argument name `time_name` are extracted into W(t).
    time_start : float
        The initial time.
    params : dict, optional
        Extra arguments to pass to the Hamiltonian of ``syst``, excluding time.

    Notes
    -----
    By definition the perturbation is defined with respect to the initial time.
    The perturbation should be therefore zero for times t < `time_start`. This
    is not inforced by this routine however, but W(t) is always evaluated
    with the time argument it gets.
    """
    cdef:
        object H
        object tparams
        int size
        int nnz
        object time_name
        object td_hamiltonian

    def __init__(self, syst, time_name, time_start, params=None):
        self.tparams = system.add_time_to_params(params, time_name, time_start)
        self.time_name = time_name
        self.H = H = syst.hamiltonian

        # Number of (potentially) non-zero entries in the time-dependent
        # Hamiltonian matrix.
        nnz = 0

        # Process onsites.
        self.td_hamiltonian = td_hamiltonian = []
        for i, (onsite, _) in enumerate(syst.onsites):
            if system.is_time_dependent_function(onsite, time_name):
                a, b = rng = system.orb_range(syst, i)
                norbs = b - a
                nnz += norbs * norbs
                td_hamiltonian.append(
                    ((i, i, rng, rng, H(i, i, params=self.tparams))))

        # Process hoppings.
        for edge_id, (i, j) in enumerate(syst.graph):
            hop_val, _ = syst.hoppings[edge_id]
            if system.is_time_dependent_function(hop_val, time_name):
                a, b = ori = system.orb_range(syst, i)
                c, d = orj = system.orb_range(syst, j)
                nnz += 2 * (b - a) * (d - c)
                td_hamiltonian.append((i, j, ori, orj, H(i, j, params=self.tparams)))

        _, _, total_norbs = syst.site_ranges[-1]
        self.nnz = nnz
        self.size = total_norbs

    @property
    def size(self):
        '''The size of the W(t) matrix.

        Returns
        -------
        size : int
            The size of the W(t) matrix is ``size`` x ``size``.
        '''
        return self.size

    @property
    def nnz(self):
        '''The total number of time-dependent matrix elements.

        Returns
        -------
        nnz : int
            The total number of time-dependent matrix elements, equal to
            len(W(t).data).
        '''
        return self.nnz

    def data(self, double time, object out=None):
        '''Return the time-dependent matrix elements.

        Parameters
        ----------
        time : int or float
            Time argument, must be equal or larger than `time_start`.
        out : `numpy.ndarray`, optional
            Data array of the W(t) matrix for operation in-place.

        Returns
        -------
        out : `numpy.ndarray` or `None`
            Data array of the W(t) matrix (W(t).data). If an ``out`` argument is given,
            the operation is performed in-place and the routine returns `None`.
        '''
        if out is None:
            do_return = True
            out = np.empty(self.nnz, complex)
        else:
            do_return = False
            assert out.shape[0] == self.nnz, _errormsg.format('output',
                                                              out.shape[0],
                                                              self.nnz)
        index = 0
        self.tparams[self.time_name] = time
        for i, j, _, _, H_0 in self.td_hamiltonian:
            H = self.H(i, j, params=self.tparams)
            mat_els = np.asarray(H - H_0).flatten()
            for val in mat_els:
                out[index] = val
                index += 1
            if i != j:
                mat_els = mat_els.conjugate()
                for val in mat_els:
                    out[index] = val
                    index += 1
        return out if do_return else None

    def row_col(self):
        '''Return the row and column vectors of the W(t) matrix.

        Returns
        -------
        row : `np.ndarray`
            Row vector of the W(t) matrix in coo format.
        col : `np.ndarray`
            Column vector of the W(t) matrix in coo format.
        '''
        row = np.empty(self.nnz, np.int32)
        col = np.empty(self.nnz, np.int32)
        index = 0
        for i, j, to, frm, H_0 in self.td_hamiltonian:
            to = range(*to)
            frm = range(*frm)
            for k in to:
                for l in frm:
                    row[index] = k
                    col[index] = l
                    index += 1
            if i != j:
                for k in to:
                    for l in frm:
                        row[index] = l
                        col[index] = k
                        index += 1
        return row, col

    def empty(self):
        '''Return an empty W(t) matrix with correct shape and indices.

        Returns
        -------
        out : `~scipy.sparse.coo_matrix`
            An "empty" W(t) matrix filled with zeros.
        '''
        data = np.zeros((self.nnz,), dtype=complex)
        row_col = self.row_col()
        return sp.coo_matrix((data, row_col), (self.size, self.size))

    def evaluate(self, double time, object out=None):
        '''Evaluate the W(t) matrix for the given time t.

        Parameters
        ----------
        time : int or float
            Time argument, must be equal or larger than `time_start`.
        out : `~scipy.sparse.coo_matrix`, optional
            Sparse matrix with to perform the operation in-place.

        Returns
        -------
        out : `~scipy.sparse.coo_matrix` or `None`
            The matrix W(t).
            The size of the W(t) matrix is ``size`` x ``size``, containing
            ``nnz`` complex entries. If an ``out`` argument is given,
            the operation is performed in-place and the routine returns `None`.
        '''
        if out is None:
            data = self.data(time)
            row_col = self.row_col()
            return sp.coo_matrix((data, row_col), (self.size, self.size))
        else:
            assert out.shape[0] == self.size, _errormsg.format('output',
                                                                out.shape[0],
                                                                self.size)
            self.data(time, out.data)

    def apply(self, double time, ket, out=None):
        '''Evaluate the matrix-vector product W(t) @ ket for the given time.

        Parameters
        ----------
        time : int or float
            Time argument, must be equal or larger than `time_start`.
        out : `numpy.ndarray`, optional
            Sparse matrix with to perform the operation in-place.
            If provided, array must be zero on input.

        Returns
        -------
        out : `numpy.ndarray` or `None`
            The product W(t) @ ket. If an ``out`` argument is given,
            the operation is performed in-place and the routine returns `None`.
        '''
        assert ket.shape[0] == self.size, _errormsg.format('Ket', ket.shape[0],
                                                           self.size)
        if out is None:
            do_return = True
            out = np.zeros((self.size,), dtype=complex)
        else:
            do_return = False
            assert out.shape[0] == self.size, _errormsg.format('output',
                                                               out.shape[0],
                                                               self.size)
        self.tparams[self.time_name] = time
        for i, j, to, frm, H_0 in self.td_hamiltonian:
            to = slice(*to)
            frm = slice(*frm)
            mat_el = self.H(i, j, params=self.tparams) - H_0
            out[to] += np.dot(mat_el, ket[frm])
            if i != j:
                out[frm] += np.dot(_herm_conj(mat_el), ket[to])
        return out if do_return else None


cdef class PerturbationInterpolator:
    r"""A class to extract and interpolate W(t).

    This class can be used like `tkwant.onebody.kernel.PerturbationExtractor`,
    to calculate W(t) and the matrix-vector product W(t) * ket.
    Interpolation is done only for performance reasons, as W(t) from
    Kwant is slow to evaluate.

    Parameters
    ----------
    syst : `kwant.builder.FiniteSystem`
        The system from which to extract the time-dependent perturbation.
    time_name : str
        The name of the time argument. Only sites with Hamiltonian value
        functions with the argument name `time_name` are extracted into W(t).
    time_start : float
        The initial time.
    params : dict, optional
        Extra arguments to pass to the Hamiltonian of ``syst``, excluding time.
    interpolation_type : optional
        The interpolator class

    Notes
    -----
    By definition the perturbation is defined with respect to the initial time.
    The perturbation should be therefore zero for times t < `time_start`. This
    is not inforced by this routine however, but W(t) is always evaluated
    with the time argument it gets.
    """

    cdef:
        double dt
        unsigned size
        unsigned nnz
        object wt
        object wfunc
        object intfunc

    def __init__(self, syst, time_name, time_start, params=None, interpolation_type=Interpolation):

        self.wfunc = PerturbationExtractor(syst, time_name, time_start, params)

        # for non-vectorized systems
        _, _, norbs = syst.site_ranges[-1]
        self.size = norbs
        self.nnz = self.wfunc.nnz

        self.intfunc = interpolation_type(self.wfunc.data, time_start)
        self.wt = self.wfunc.empty()

    @property
    def size(self):
        '''The size of the W(t) matrix.

        Returns
        -------
        size : int
            The size of the W(t) matrix is ``size`` x ``size``.
        '''
        return self.size

    @property
    def nnz(self):
        '''The total number of time-dependent matrix elements.

        Returns
        -------
        nnz : int
            The total number of time-dependent matrix elements, equal to
            len(W(t).data).
        '''
        return self.nnz

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def evaluate(self, double time, object out=None):
        '''Evaluate the W(t) matrix for the given time t.

        Parameters
        ----------
        time : int or float
            Time argument, must be equal or larger than `time_start`.
        out : `~scipy.sparse.coo_matrix`, optional
            Sparse matrix with to perform the operation in-place.

        Returns
        -------
        out : `~scipy.sparse.coo_matrix` or `None`
            The matrix W(t).
            The size of the W(t) matrix is ``size`` x ``size``, containing
            ``nnz`` complex entries. If an ``out`` argument is given,
            the operation is performed in-place and the routine returns `None`.
        '''

        if out is None:
            self.intfunc.eval(time, self.wt.data)
            return self.wt
        else:
            assert out.shape[0] == self.size, _errormsg.format('output',
                                                               out.shape[0],
                                                               self.size)
            assert out.data.shape[0] == self.nnz, _errormsg.format('output.dat',
                                                                   out.data.shape[0],
                                                                   self.nnz)
            self.intfunc.eval(time, out.data)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def apply(self, double time, const complex[:] ket, object out=None):
        '''Evaluate the matrix-vector product W(t) @ ket for the given time.

        Parameters
        ----------
        time : int or float
            Time argument, must be equal or larger than `time_start`.
        out : `numpy.ndarray`, optional
            Sparse matrix with to perform the operation in-place.

        Returns
        -------
        out : `numpy.ndarray` or `None`
            The product W(t) @ ket. If an ``out`` argument is given,
            the operation is performed in-place and the routine returns `None`.
        '''

        assert ket.shape[0] == self.size, _errormsg.format('Ket',
                                                           ket.shape[0],
                                                           self.size)
        if out is None:
            do_return = True
            out = np.zeros((self.size,), dtype=complex)
        else:
            assert out.shape[0] == self.size, _errormsg.format('output',
                                                               out.shape[0],
                                                               self.size)

        self.evaluate(time, self.wt)

        coo_matvec(
            self.wt.nnz, self.wt.row, self.wt.col, self.wt.data,
            ket, out)

        return out if do_return else None


# Example Kernel implementation in C

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef _csr_spmv_ptr(const int size, const int *indptr, const int *indices,
                   const complex *data,  const complex *X, complex *Y):
    """Naive implementation of CSR-format matrix-vector multiplication.

    This is taken almost verbatim from scipy.sparse.
    """
    cdef complex sum
    cdef unsigned i, jj

    for i in range(size):
       sum = Y[i]
       for jj in range(indptr[i], indptr[i+1]):
           sum += data[jj] * X[indices[jj]]
       Y[i] = sum


cdef class Simple(Kernel):
    """A C-implemented kernel using naive CSR matvec."""

    cdef:
        object H0
        object W
#        PerturbationExtractor W
        int center_size
        complex[:] psi_st
        object _temp

    def __init__(self, H0, W, complex[:] psi_st=None, work=None):
        pass

    def __cinit__(self, H0, W, complex[:] psi_st=None, work=None):

        if not isinstance(H0, sp.csr.csr_matrix):
            self.H0 = H0.tocsr()  # self.H0 is no reference any more in that case
        else:
            self.H0 = H0

        self.W = W
        self.size = H0.shape[0]

        if W is not None:
            self.center_size = self.W.size
        else:
            self.center_size = self.size
        self.psi_st = psi_st
        if work is None:
            self._temp = np.empty((self.center_size,), dtype=complex)
        else:
            if work.size < self.center_size:
                raise ValueError('work array size={} must be at least '
                     'the size of the central system={}'
                     .format(work.size, self.center_size))
            self._temp = work


    def rhs(self, energy=None):

        if energy is None:
            H0 = self.H0
        else:
            H0 = self.H0 - energy * sp.eye(self.size)

        # The approach with a separated class ``_CalcRHS`` to calculate the r.h.s.
        # seems more complicated, but it allows to store the static Hamiltonian
        # as a reference in class ``Scipy(Kernel)`` in ``self.H0``
        # (for manybody calculations there is onle reference per rank).
        # Moreover, the calculation ``H = H0 - E`` is done inplace,
        # and only once when the Scipy(Kernel).rhs() method is called.
        # It is important to precalculate H0 - E at this stage
        # (typically when ``WaveFunction.evolve()`` is called) and not
        # to do the inplace calculation in ``_CalcRHS.__call__()``,
        # as the later method is called very often (by the DGL solver).
        # The implace calculation generates only a temporary object ``H``
        # which not stored in the ``onebody.WaveFunction``, such
        # that it is cleaned by the garbage collector.
        return _CalcRHSc(H0.indptr, H0.indices, H0.data, self.W, self.psi_st, 
                         self.size, self.center_size, self._temp)

    @property
    def W(self):
        '''The time dependent perturbation W(t)'''
        return self.W

    def set_W(self, W):
        '''Set the time dependent perturbation W(t)'''
        self.W = W


cdef class _CalcRHSc:

    cdef:
        object W
        #PerturbationExtractor W
        int center_size, size
        complex *psi_st
        complex *temp
        complex[:] _data
        int[:] _indices, _indptr

    def __init__(self, int[:] indptr, int[:] indices, complex[:] data, W, complex[:] psi_st, 
                 int size, int center_size, complex[:] temp):
        pass

    """Calculate the right-hand-side of the time-dependent Schrödinger equation."""
    def __cinit__(self, int[:] indptr, int[:] indices, complex[:] data, W, complex[:] psi_st, 
                  int size, int center_size, complex[:] temp):

        self._data = data
        self._indices = indices
        self._indptr = indptr
        self.W = W
        if psi_st is None:
            self.psi_st = NULL
        else:
            self.psi_st = &psi_st[0]
        self.size = size
        self.center_size = center_size

        # the id of the temp array is not the same any more due to argument complex[:] temp
        #print('simple kernel work', id(temp))
        #exit(0)

        self.temp = &temp[0]


    # RHS implemented as a static method to conform to the solver interface.
    @cython.boundscheck(False)
    @cython.wraparound(False)
    def __call__(self, const complex[::1] psi, complex[::1] dpsidt, double time):

        cdef int *indptr
        cdef int *indices
        cdef complex *data

        data = &self._data[0]
        indices = &self._indices[0]
        indptr = &self._indptr[0]

        cdef unsigned i
        for i in range(self.size):
            dpsidt[i] = 0

        ### H0 @ psi -> dpsidt
        _csr_spmv_ptr(self.size, indptr, indices, data, &psi[0], &dpsidt[0])


        ### W(t) @ psi + dpsidt -> dpsidt
        if self.psi_st == NULL:
            try:
                self.W.apply(time, <complex[:self.center_size]>&psi[0],
                             out=<complex[:self.center_size]>&dpsidt[0])
            except Exception as exept:
                if self.W is None:
                    pass
                else:
                    raise exept
        else:
            # we need to act on `temp = psi + psi_st`
            # use pre-allocated storage
            for i in range(self.center_size):
                self.temp[i] = psi[i] + self.psi_st[i]
            try:
                self.W.apply(time, <complex[:self.center_size]>self.temp,
                             out=<complex[:self.center_size]>&dpsidt[0])
            except Exception as exept:
                if self.W is None:
                    pass
                else:
                    raise exept
        ### dpsidt = -1j * dpsidt
        for i in range(self.size):
            dpsidt[i] = -1j * dpsidt[i]

    def __reduce__(self):
        raise NotImplementedError("use the scipy kernel for serialization")

default = Simple



# Script 1/3 to produce the figure in Tkwant's tutorial
# "Self-consistent Tkwant: a generic solver for time-dependent mean field calculations"
# in the section on Luttinger liquids.
#
# Step 1: run the two simulation (plasmon_u_0_computation.py, plasmon_u_10_computation.py)
#         preferably on a cluster using MPI
# Step 2: plot the data (plasmon_plot_results.py)
#
import tkwant
import kwant
import kwantspectrum

import functools
import numpy as np
import scipy
import cmath
import sys
import pickle

import logging
tkwant.logging.level = logging.INFO  # possible levels (increasing verbosity): WARNING, INFO, DEBUG

# --- MPI helper routines


def am_master():
    """Return true for the MPI master rank"""
    return tkwant.mpi.get_communicator().rank == 0


def print_master(*args, **kwargs):
    if am_master():
        print(*args, **kwargs)
    sys.stdout.flush()


# --- Non-interacting Kwant system

def phi(time):
    t0 = 200
    A = 0.1 * np.pi
    sigma = 60.056120439
    return A * (1 + scipy.special.erf((time - t0) / sigma))


# time dependent coupling with gaussian pulse
def gamma_t(site1, site2, time):
    return - cmath.exp(- 1j * phi(time))


def make_system(L=3000):

    # system building
    lat = kwant.lattice.square(a=1, norbs=1)
    syst = kwant.Builder()

    # central scattering region
    syst[(lat(x, 0) for x in range(-L, L))] = 0
    syst[lat.neighbors()] = -1
    syst[lat(0, 0), lat(-1, 0)] = gamma_t

    # add leads
    sym = kwant.TranslationalSymmetry((-1, 0))
    lead_left = kwant.Builder(sym)
    lead_left[lat(0, 0)] = 0
    lead_left[lat.neighbors()] = -1
    syst.attach_lead(lead_left)
    syst.attach_lead(lead_left.reversed())

    return syst


# --- Interaction

class Potential:
    def prepare(self, ft, tmax):
        self._ft = ft

    def evaluate(self, time):
        diag = self._ft(time)
        return scipy.sparse.diags([diag], [0], dtype=complex)


class HartreeOperator:
    """An operator to calculate the Hartree potential Q(t) = U (n(t) - n(0))"""
    def __init__(self, density_operator, interaction_strength, density0):
        self._density_operator = density_operator
        self._interaction_strength = interaction_strength
        self._density0 = density0

    def __call__(self, bra, ket=None, args=(), *, params=None):
        """API similar to a Kwant operator"""
        density = self._density_operator(bra, ket, args, params=params)
        return (density - self._density0) * self._interaction_strength


syst = make_system().finalized()

sites = [site.pos[0] for site in syst.sites]
times = [500, 700, 900, 1100]
u = 0  # interaction strength

# calculate the spectrum E(k) for all leads
spectra = kwantspectrum.spectra(syst.leads)

# set the chemical potential in the leads
occupations = tkwant.manybody.lead_occupation(chemical_potential=-1)
emin, emax = tkwant.manybody.calc_energy_cutoffs(occupations)

# define boundary conditions
bdr = tkwant.leads.automatic_boundary(spectra, tmax=max(times), emin=emin, emax=emax)

# calculate the k intervals for the quadrature
interval_type = functools.partial(tkwant.manybody.Interval, order=20,
                                  quadrature='gausslegendre')
intervals = tkwant.manybody.calc_intervals(spectra, occupations, interval_type)
intervals = tkwant.manybody.split_intervals(intervals, number_subintervals=60)

# calculate all onebody scattering states at t = 0
tasks = tkwant.manybody.calc_tasks(intervals, spectra, occupations)
psi_init = tkwant.manybody.calc_initial_state(syst, tasks, bdr)

# set up the non-interacting manybody wave function
wave_function = tkwant.manybody.WaveFunction(psi_init, tasks)

density_operator = kwant.operator.Density(syst)

density0 = wave_function.evaluate(density_operator)

densities = []
for time in times:
    print_master('time= {}'.format(time))
    wave_function.evolve(time)
    density = wave_function.evaluate(density_operator)
    if am_master():
        density -= density0
    densities.append(density)

results = {}
results['densities'] = densities
results['times'] = times
results['sites'] = sites

#  write the result into a file
if am_master():
    filename = 'plasmon_u_{}_data.npy'.format(u)
    pickle.dump(results, open(filename, "wb"))



# Script 1/2 to produce the figure in Tkwant's tutorial
# "Self-consistent Tkwant: a generic solver for time-dependent mean field calculations"
# in the section on Microscopic Bogoliubov-deGennes equations.
#
# Step 1: run this script (preferably on a cluster using MPI parallelization)
# Step 2: plot the results by running the script bdg_junction_plot_results.py
#
import tkwant
import kwant
import kwantspectrum

import cmath
import numpy as np
import scipy.integrate
import scipy.interpolate
import functools
import tinyarray
import pickle


comm = tkwant.mpi.get_communicator()

sx = tinyarray.array([[0, 1], [1, 0]], complex)
sz = tinyarray.array([[1, 0], [0, -1]], complex)
sa = tinyarray.array([[1, 0], [0, 0]], complex)
sb = tinyarray.array([[0, 0], [0, 1]], complex)

# parameters for pulse
tmax = 1000
tau = 100
delta = 0.1
U = 2
Ef = 0
w0 = delta
dV = w0 / (2 * tmax - tau)
taupi2 = tau**2 / np.pi**2


def V0(t):
    if t < tau:
        return dV * (t - np.sin(t / tau * np.pi) * tau / np.pi)
    else:
        return 2 * dV * (t - 0.5 * tau)


def phi0(t):
    if t < tau:
        return dV * (0.5 * t**2 + (np.cos(t / tau * np.pi) - 1) * taupi2)
    else:
        return dV * (t**2 - tau * t + 0.5 * tau**2 - taupi2)


def make_SNS_system(delta, U, Ef):

    # system building
    lat = kwant.lattice.square(a=1, norbs=2)
    syst = kwant.Builder()

    # central scattering region
    onsite_N = (U - Ef) * sz
    onsite_S = - Ef * sz + delta * sx
    syst[lat(-1, 0)] = onsite_S
    syst[lat(0, 0)] = onsite_N
    syst[lat(1, 0)] = onsite_S
    syst[lat.neighbors()] = sz

    # add leads
    lead_left = kwant.Builder(kwant.TranslationalSymmetry((-1, 0)))
    lead_left[lat(0, 0)] = onsite_S
    lead_left[lat.neighbors()] = sz
    syst.attach_lead(lead_left)
    syst.attach_lead(lead_left.reversed())

    return syst, lat


class BdGPotential:
    """A class to calculate Q(t) for the Bogoliubov-deGennes equations"""

    def __init__(self, syst, lat, w0, q, r):

        self._a = w0 / q
        self._b = w0**2
        self._c = r * w0 / q
        self._phi_init = 0
        self._v_init = 0
        self._tmin = 0
        self._tmax = 0
        self.time = []
        self.current = []
        self.phi = []
        self.v = []

        # Prepare a self-consistent hamiltonian matrix Q(t) with non-zero
        # 2x2 submatrix between site -1 and 0
        self._qt = tkwant.system.Hamiltonian(syst, hopping=(lat(0, 0), lat(-1, 0)))

    def prepare(self, current_func, tmax):
        """Pre-calculate the interaction contribution Q(t)"""
        self._tmin = self._tmax
        self._tmax = tmax

        # time grid for the solution/interpolation of the BdG differential equation
        times = np.linspace(self._tmin, self._tmax, num=4)

        def calc_rhs(tt, yy):  # right-hand-side of the BdG differential equation
            phi, v = yy
            return [v, - self._a * v - self._b * phi + self._c * current_func(tt)]

        # solve the BdG differential equation: d(phi, V) / dt = rhs
        dgl = scipy.integrate.ode(calc_rhs).set_integrator('dopri5')
        dgl.set_initial_value([self._phi_init, self._v_init], self._tmin)

        phi = [self._phi_init]
        for time in times[1:]:
            result = dgl.integrate(time)
            assert dgl.successful(), 'ode integration problem'
            phi.append(result[0])

        # save I(t), phi(t), V(t)
        # which are current, phase and voltage trought the classical circuit
        self.time.append(self._tmin)
        self.current.append(current_func(self._tmin))
        self.phi.append(self._phi_init)
        self.v.append(self._v_init)

        # initial values for the next update step
        self._phi_init = result[0]
        self._v_init = result[1]

        # interpolate phi(t) for t in [tmin, tmax]
        self._phi_func = scipy.interpolate.interp1d(times, phi, kind='cubic')

    def evaluate(self, time):
        """Return the interaction contribution Q(t) evaluated at time t in [tmin, tmax]"""
        phi_j = phi0(time) - self._phi_func(time)  # phi_j is the phase trough the junction
        ephi = cmath.exp(- 1j * phi_j) - 1
        qmat = ephi * sa - ephi.conjugate() * sb  # subblock of self-consistent matrix Q(t)
        return self._qt.get(qmat)


def main():

    # initialize the tight-binding system
    syst, lat = make_SNS_system(delta, U, Ef)
    syst = syst.finalized()

    # set chemical potential and zero temperature (default)
    occupations = tkwant.manybody.lead_occupation(chemical_potential=Ef)

    # calculate the spectrum E(k) for all leads
    spectra = kwantspectrum.spectra(syst.leads)

    # define boundary conditions, set upper cutoff energy to Ef
    boundaries = tkwant.leads.automatic_boundary(spectra, tmax, emax=Ef)

    # calculate the k intervals for the quadrature
    interval_type = functools.partial(tkwant.manybody.Interval, order=20,
                                      quadrature='gausslegendre')
    intervals = tkwant.manybody.calc_intervals(spectra, occupations, interval_type)
    intervals = tkwant.manybody.split_intervals(intervals, number_subintervals=40)

    # calculate all onebody scattering states at t = 0
    tasks = tkwant.manybody.calc_tasks(intervals, spectra, occupations)
    psi_init = tkwant.manybody.calc_initial_state(syst, tasks, boundaries)

    # set up the manybody wave function
    wave_function = tkwant.manybody.WaveFunction(psi_init, tasks)

    # measure the current from the center (site 0) to the right lead (site 1)
    current_operator = kwant.operator.Current(syst, onsite=sz,
                                              where=[(lat(1, 0), lat(0, 0))])

    # self-consistent potential
    bdg_potential = BdGPotential(syst, lat, w0=w0, q=20, r=3*np.sqrt(2))

    # set up the self-consistent interacting manybody wave function
    sc_wavefunc = tkwant.interaction.SelfConsistentState(wave_function,
                                                         current_operator,
                                                         bdg_potential)

    # evolve the interacting wave function up to tmax
    sc_wavefunc.evolve(tmax)

    # save the result
    if comm.rank == 0:

        v0t = [V0(t) for t in bdg_potential.time]

        results = {}
        results['times'] = np.array(bdg_potential.time)
        results['current'] = np.array(bdg_potential.current)
        results['phi'] = np.array(bdg_potential.phi)
        results['v'] = np.array(bdg_potential.v)
        results['v0t'] = np.array(v0t)
        results['delta'] = delta
        pickle.dump(results, open('self_consistent_bdg_data.npy', "wb"))


if __name__ == '__main__':
    main()



import numpy as np
import matplotlib
from matplotlib import pyplot as plt

import kwant
import tkwant


def make_system(a=1, gamma=1.0, radius=10, width=7):
    """Make a tight binding system on a single square lattice"""
    # `a` is the lattice constant and `gamma` the hopping integral
    # both set by default to 1 for simplicity.

    lat = kwant.lattice.square(a, norbs=1)
    syst = kwant.Builder()

    # Define the quantum dot
    def circle(pos):
        (x, y) = pos
        return x ** 2 + y ** 2 < radius ** 2

    syst[lat.shape(circle, (0, 0))] = 4 * gamma
    syst[lat.neighbors()] = -gamma

    lead = kwant.Builder(kwant.TranslationalSymmetry((-1, 0)))
    lead[(lat(0, j) for j in range(-(width - 1) // 2, width // 2 + 1))] = 4 * gamma
    lead[lat.neighbors()] = -gamma
    syst.attach_lead(lead)
    syst.attach_lead(lead.reversed())

    return syst


def faraday_flux(time):
    return 0.05 * (time - 10 * np.sin(0.1 * time))


def main():

    # create a system andadd a time-dependent voltage to lead 0 -- this is
    # implemented by adding sites to the system at the interface with the lead
    # and multiplying the hoppings to these sites by exp(-1j * faraday_flux(time))
    syst = make_system()

    added_sites = tkwant.leads.add_voltage(syst, 0, faraday_flux)
    interface_hoppings = [(a, b)
                          for b in added_sites
                          for a in syst.neighbors(b) if a not in added_sites]

    fsyst = syst.finalized()

    current_operator = kwant.operator.Current(fsyst, where=interface_hoppings,
                                              sum=True)

    times = np.arange(0, 600)

    # create a time-dependent wavefunction that starts in a scattering state
    # originating from the left lead
    mode = 0
    psi = tkwant.onebody.ScatteringStates(fsyst, energy=1., lead=0,
                                          tmax=max(times))[mode]

    # evolve forward in time, calculating the current
    current = []
    for time in times:
        psi.evolve(time)
        current.append(psi.evaluate(current_operator))

    # plot results
    plt.figure(figsize=(18, 7))

    gs = matplotlib.gridspec.GridSpec(2, 2)

    ax1 = plt.subplot(gs[0, 0])
    ax1.plot(times, current)
    ax1.set_ylabel(r'current $I(t)$')

    ax2 = plt.subplot(gs[1, 0])
    ax2.plot(times, 0.05 * (1 - np.cos(0.1 * times)))
    ax2.set_xlabel(r'time $t$')
    ax2.set_ylabel(r'voltage $V(t)$')

    ax3 = plt.subplot(gs[:, 1])
    kwant.plot(syst, ax=ax3, site_color='k', lead_color='grey', num_lead_cells=4,
               hop_lw=lambda a, b: 0.3 if (a, b) in interface_hoppings else 0.1,
               hop_color=lambda a, b: 'r' if (a, b) in interface_hoppings else 'k')
    plt.show()

if __name__ == '__main__':
    main()



import tkwant
import kwant

import numpy as np
import scipy.special
import scipy.integrate
import cmath
import functools as ft
import matplotlib.pyplot as plt

import sys
import pickle

import logging
tkwant.logging.level = logging.INFO  # possible levels (increasing verbosity): WARNING, INFO, DEBUG

comm = tkwant.mpi.get_communicator()


def am_master():
    return comm.rank == 0


def print_master(*args, **kwargs):
    if am_master():
        print(*args, **kwargs)
    sys.stdout.flush()

# parameters
gamma = 0.1  # coupling
epsilon = 0.1  # onsite element
time0 = 100

times = np.linspace(0, 200, 201)


def gamma_t(site1, site2, time):
    # time dependent coupling function \gamma(t) = \Theta(t) * gamma
    if time <= 0:
        return 0
    return - gamma


def make_siam_system(gt, epsilon):

    # system building
    lat = kwant.lattice.square(a=1, norbs=1)
    syst = kwant.Builder()

    # central scattering region
    syst[(lat(x, 0) for x in [-1, 0, 1])] = 0
    syst[lat(0, 0)] = epsilon

    syst[(lat(0, 0), lat(-1, 0))] = gt
    syst[(lat(1, 0), lat(0, 0))] = gt

    # add leads
    sym = kwant.TranslationalSymmetry((-1, 0))
    lead_left = kwant.Builder(sym)
    lead_left[lat(0, 0)] = 0
    lead_left[lat.neighbors()] = -1
    syst.attach_lead(lead_left)
    syst.attach_lead(lead_left.reversed())

    return syst, lat

# Switch off perturbation interpolation for sudden switch
onebody_wavefunction_type = ft.partial(tkwant.onebody.WaveFunction.from_kwant,
                                       perturbation_type=tkwant.onebody.kernels.PerturbationExtractor)

scattering_state_type = ft.partial(tkwant.onebody.ScatteringStates,
                                   wavefunction_type=onebody_wavefunction_type)

# --- Tkwant simulation for the dynamics of G^<(t, t0)

syst, lat = make_siam_system(gamma_t, epsilon)
syst = syst.finalized()

green = tkwant.manybody.GreenFunction(syst, max(times),
                                      scattering_state_type=scattering_state_type)

green_less = []

for time in times:

    print_master('time= {}'.format(time))

    green.evolve(time, time0)
    green.refine_intervals()
    green_less.append(green.lesser(1, 1))


# --- Tkwant simulation for the dynamics G^<(t, t)

density_operator = kwant.operator.Density(syst, where=[lat(0, 0)])

green = tkwant.manybody.GreenFunction(syst, max(times),
                                      scattering_state_type=scattering_state_type)

state = tkwant.manybody.State(syst, max(times),
                              scattering_state_type=scattering_state_type)

green_less_tt = []
nt = []

for time in times:

    print_master('time= {}'.format(time))

    green.evolve(time, time)
    green.refine_intervals()
    green_less_tt.append(green.lesser(1, 1))

    state.evolve(time)
    state.refine_intervals()
    nt.append(state.evaluate(density_operator))


# --- Tkwant simulation for the equilibrium density

syst_eq, lat = make_siam_system(gamma, epsilon)
syst_eq = syst_eq.finalized()

density_operator = kwant.operator.Density(syst_eq, where=[lat(0, 0)])

state = tkwant.manybody.State(syst_eq, tmax=1)
dens_tkwant_equilibrium = state.evaluate(density_operator)


# --------. Analytica formulas

def n_equilibrium(gamma, epsilon=0):
    # siam equlibrium density on the dot in flatband approx for mu_L = mu = R = 0
    Gamma = 2 * gamma**2
    return np.arctan(- epsilon / Gamma) / np.pi + 0.5


def integrate_quad(func, a, b):

    def func_re(time):
        return func(time).real

    def func_im(time):
        return func(time).imag

    epsabs = 1.49e-8
    epsrel = 1.49e-8
    limit = 5000
    return (scipy.integrate.quad(func_re, a, b, epsabs=epsabs, epsrel=epsrel, limit=limit)[0]
            + 1j * scipy.integrate.quad(func_im, a, b, epsabs=epsabs, epsrel=epsrel, limit=limit)[0])


def green_lesser_flatband(t0, t1, gamma, epsilon=0):
    # siam G^< on the dot in flatband approx for mu_L = mu = R = 0 after sudden switching of the leads
    # using analytical fourier tranform
    if t0 <= 0 or t1 <= 0:
        return 0
    Gamma = 2 * gamma**2
    iless = tkwant.special._iless  # I^<_t(a) =  - \int_{-\infty}^{0} d \omega \frac{e^{- i \omega t}}{\omega - a}

    def it(t):
        return iless(epsilon - 1j * Gamma, t) - iless(epsilon + 1j * Gamma, t)

    neq = n_equilibrium(gamma, epsilon)
    if t0 == t1:
        return (1j * neq * (1 + np.exp(- 2 * Gamma * t0)) - 0.5 / np.pi * np.exp(- Gamma * t0)
                * (cmath.exp(1j * epsilon * t0) * it(t0) + cmath.exp(-1j * epsilon * t0) * it(-t0)))
    return 0.5 / np.pi * (it(t0 - t1)
                          + 2 * np.pi * 1j * cmath.exp(-1j * epsilon * (t0 - t1) - Gamma * (t0 + t1)) * neq
                          - cmath.exp(1j * epsilon * t1 - Gamma * t1) * it(t0)
                          - cmath.exp(-1j * epsilon * t0 - Gamma * t0) * it(-t1))


def green_lesser_flatband_num(t0, t1, gamma, epsilon=0):
    # siam G^< on the dot in flatband approx for mu_L = mu = R = 0 after sudden switching of the leads
    # using numerical fourier transform
    if t0 <= 0 or t1 <= 0:
        return 0
    Gamma = 2 * gamma**2

    def integrand_n(w):
        y = w - epsilon
        return (cmath.exp(-1j * t0 * y + Gamma * t0) - 1) * (cmath.exp(1j * t1 * y + Gamma * t1) - 1) / (Gamma**2 + y**2)
    fac = 1j * Gamma / np.pi * cmath.exp(-1j * epsilon * (t0 - t1)) * np.exp(- Gamma * (t0 + t1))
    return fac * integrate_quad(integrand_n, -100, 0)


def green_lesser_riwar(t0, t1, gamma, epsilon=0):
    # this is the form of the paper Eq. 15
    # R.-P. Riwar and T. L. Schmidt, “Transient dynamics of a molecular quantum dot with a vibrational degree of freedom,” Phys. Rev. B 80,125109 (2009).
    if t0 <= 0 or t1 <= 0:
        return 0
    Gamma = 2 * gamma**2

    def integrand_n(w):
        y = w - epsilon
        return (cmath.exp(-1j * t0 * y) - np.exp(- Gamma * t0)) * (cmath.exp(1j * t1 * y) - np.exp(- Gamma * t1)) / (Gamma**2 + y**2)
    fac = 1j * Gamma / np.pi
    return fac * integrate_quad(integrand_n, -100, 0)


dens_flatband_equilibrium = n_equilibrium(gamma, epsilon)
gless_tt_flatband = np.array([green_lesser_flatband(t, t, gamma, epsilon) for t in times])
gless_tt0_flatband = np.array([green_lesser_flatband(t, time0, gamma, epsilon) for t in times])
gless_tt_flatband_num = np.array([green_lesser_flatband_num(t, t, gamma, epsilon) for t in times])
gless_tt0_flatband_num = np.array([green_lesser_flatband_num(t, time0, gamma, epsilon) for t in times])
gless_tt_riwar = np.array([green_lesser_riwar(t, t, gamma, epsilon) for t in times])
gless_tt0_riwar = np.array([green_lesser_riwar(t, time0, gamma, epsilon) for t in times])


# --------------- Store result

results = {}
results['times'] = times
results['gamma'] = gamma
results['epsilon'] = epsilon
results['t0'] = time0

results['gless_tt0_tkwant'] = np.array(green_less)
results['gless_tt_tkwant'] = np.array(green_less_tt)
results['n_t_tkwant'] = np.array(nt)
results['dens_tkwant_equilibrium'] = dens_tkwant_equilibrium

results['gless_tt_flatband'] = gless_tt_flatband
results['gless_tt0_flatband'] = gless_tt0_flatband
results['gless_tt_flatband_num'] = gless_tt_flatband_num
results['gless_tt0_flatband_num'] = gless_tt0_flatband_num
results['dens_flatband_equilibrium'] = dens_flatband_equilibrium

results['gless_tt_riwar'] = gless_tt_riwar
results['gless_tt0_riwar'] = gless_tt0_riwar

if am_master():
    pickle.dump(results, open('siam_sudden_switch_data.npy', "wb"))
